from __future__ import annotations

import typing as t

from datasets import Dataset

from centricube_ragas import evaluate as ragas_evaluate
from centricube_ragas.evaluation import Result
from centricube_ragas.metrics.base import Metric

if t.TYPE_CHECKING:
    from llama_index.indices.query.base import BaseQueryEngine


def evaluate(
    query_engine: BaseQueryEngine,
    metrics: list[Metric],
    questions: list[str],
    ground_truths: t.Optional[list[str]] = None,
) -> Result:
    """
    Run evaluation of llama_index QueryEngine with different metrics

    Parameters
    ----------
    query_engine : BaseQueryEngine
        The QueryEngine that is to be evaluated
    metrics : list[Metric]
        The ragas metrics to use for evaluation.
    questions : list[str]
        List of questions to evaluate on
    ground_truths : list[str], optional
        List of ground_truths answer to the question to evaluate on.

    Returns
    -------
    Result
        Result object containing the scores of each metric. You can use this do analysis
        later. If the top 3 metrics are provided then it also returns the `ragas_score`
        for the entire pipeline.

    Raises
    ------
    ValueError
        if validation fails because the columns required for the metrics are missing or
        if the columns are of the wrong format.

    Examples
    --------
    Once you have a llama_index QueryEngine created you can use it to evaluate on a list
    of questions.

    Import everything you need:

    >>> from centricube_ragas.metrics import faithfulness, answer_relevancy, context_precision
    >>> from centricube_ragas.metrics.critique import harmfulness
    >>> from centricube_ragas.llama_index import evaluate

    init the query engine, get the questions and choose the metrics you want to use:

    >>> query_engine = # from llamaindex
    >>> questions: list[str] = [] # from somewhere
    >>> metrics = [faithfulness, answer_relevancy, context_precision, harmfulness]

    Run the evaluation:

    >>> r = evaluate(query_engine, metrics, questions)

    analysis the result:

    >>> print(r) # prints the scores of each metric
    >>> r.to_pandas() # returns a pandas dataframe if you want to do further analysis
    """

    try:
        from llama_index.async_utils import run_async_tasks
    except ImportError:
        raise ImportError(
            "llama_index must be installed to use this function. "
            "Please, install it with `pip install llama_index`."
        )

    # TODO: rate limit, error handling, retries
    responses = run_async_tasks([query_engine.aquery(q) for q in questions])

    answers = []
    contexts = []
    for r in responses:
        answers.append(r.response)
        contexts.append([c.node.get_content() for c in r.source_nodes])
    dataset_dict = {
        "question": questions,
        "answer": answers,
        "contexts": contexts,
    }
    if ground_truths is not None:
        dataset_dict["ground_truths"] = ground_truths
    ds = Dataset.from_dict(dataset_dict)
    result = ragas_evaluate(ds, metrics)

    return result



from __future__ import annotations

import typing as t
from collections import defaultdict

from datasets import Dataset
from langchain.callbacks.manager import CallbackManagerForChainRun
from langchain.chains.base import Chain
from langchain.schema import RUN_KEY
from langsmith.evaluation import EvaluationResult, RunEvaluator
from langsmith.schemas import Example, Run

from centricube_ragas.metrics.base import EvaluationMode, Metric
from centricube_ragas.validation import EVALMODE_TO_COLUMNS

if t.TYPE_CHECKING:
    from langchain.callbacks.base import Callbacks


class RagasEvaluatorChain(Chain, RunEvaluator):
    """
    Wrapper around ragas Metrics to use them with langsmith.
    """

    metric: Metric

    def __init__(self, **kwargs: t.Any):
        super().__init__(**kwargs)
        self.metric.init_model()

    @property
    def input_keys(self) -> list[str]:
        keys = ["query", "result"]
        if self.metric.evaluation_mode in [EvaluationMode.qac, EvaluationMode.qc]:
            keys += ["source_documents"]
        if self.metric.evaluation_mode in [EvaluationMode.gc]:
            keys += ["ground_truths"]
        return keys

    @property
    def output_keys(self) -> list[str]:
        return [f"{self.metric.name}_score"]

    def _call(
        self,
        inputs: dict[str, t.Any],
        run_manager: t.Optional[CallbackManagerForChainRun] = None,
    ) -> dict[str, t.Any]:
        """
        Call the evaluation chain.
        """
        self._validate(inputs)
        contexts = []

        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        callbacks = _run_manager.get_child()

        if "source_documents" in inputs:
            for document in inputs["source_documents"]:
                if isinstance(document, dict):
                    contexts.append(document["page_content"])
                else:
                    contexts.append(document.page_content)
        ground_truths = []
        if "ground_truths" in inputs:
            ground_truths = inputs["ground_truths"]

        question = inputs["query"]
        answer = inputs["result"]
        score = self.metric.score_single(
            {
                "question": question,
                "answer": answer,
                "contexts": contexts,
                "ground_truths": ground_truths,
            },
            callbacks=callbacks,
        )
        return {f"{self.metric.name}_score": score}

    def _validate(
        self,
        input: dict[str, t.Any],
        question_key: str = "query",
        prediction_key: str = "result",
        context_key: str = "source_documents",
    ) -> None:
        ...
        # validate each example
        required_columns = EVALMODE_TO_COLUMNS[self.metric.evaluation_mode]
        if "question" in required_columns and question_key not in input:
            raise ValueError(
                f'"{question_key}" is required in each example'
                f"for the metric[{self.metric.name}] you have chosen."
            )
        if "answer" in required_columns and prediction_key not in input:
            raise ValueError(
                f'"{prediction_key}" is required in each prediction'
                f"for the metric[{self.metric.name}] you have chosen."
            )
        if "contexts" in required_columns and context_key not in input:
            raise ValueError(
                f'"{context_key}" is required in each prediction for the '
                f"metric[{self.metric.name}] you have chosen."
            )
        if "ground_truths" in required_columns and "ground_truths" not in input:
            raise ValueError(
                f'"ground_truths" is required in each prediction for the '
                f"metric[{self.metric.name}] you have chosen."
            )

    def evaluate(
        self,
        examples: t.Sequence[dict],
        predictions: t.Sequence[dict],
        question_key: str = "query",
        prediction_key: str = "result",
        context_key: str = "source_documents",
        ground_truths_key: str = "ground_truths",
        *,
        callbacks: Callbacks = None,
    ) -> list[dict]:
        """Evaluate question answering examples and predictions."""
        dataset_dict = defaultdict(list)

        # validation
        if len(examples) != len(predictions):
            raise ValueError(
                "number of examples and predictions must be same. Got "
                f"len(examples)={len(examples)} and len(predictions)={len(predictions)}"
            )

        for i, example in enumerate(examples):
            self._validate(
                {**example, **predictions[i]}, question_key, prediction_key, context_key
            )
            # transform into Dataset that is supported by ragas
            if self.metric.evaluation_mode in [
                EvaluationMode.qac,
                EvaluationMode.qc,
                EvaluationMode.qa,
            ]:
                dataset_dict["question"].append(example[question_key])

            if self.metric.evaluation_mode in [EvaluationMode.qac, EvaluationMode.qa]:
                dataset_dict["answer"].append(predictions[i][prediction_key])

            if self.metric.evaluation_mode in [
                EvaluationMode.qac,
                EvaluationMode.qc,
                EvaluationMode.gc,
            ]:
                dataset_dict["contexts"].append(
                    [d.page_content for d in predictions[i][context_key]]
                )

            if self.metric.evaluation_mode == EvaluationMode.gc:
                if isinstance(example["ground_truths"], list):
                    dataset_dict["ground_truths"].append(example["ground_truths"])
                else:
                    dataset_dict["ground_truths"].append([example["ground_truths"]])

        dataset = Dataset.from_dict(dataset_dict)

        # evaluate
        dataset_with_scores = self.metric.score(dataset, callbacks=callbacks)
        scores = [
            {f"{self.metric.name}_score": score}
            for score in dataset_with_scores[self.metric.name]
        ]
        return scores

    def evaluate_run(
        self, run: Run, example: t.Optional[Example] = None
    ) -> EvaluationResult:
        """
        Evaluate a langsmith run
        """
        if run.outputs is None:
            raise ValueError("The chain should return results and service_document.")
        if example is None:
            raise ValueError("Examples have to be provided.")
        chain_eval = run.outputs
        chain_eval["query"] = run.inputs["query"]
        if self.metric.evaluation_mode == EvaluationMode.gc:
            if example.outputs is None or "ground_truths" not in example.outputs:
                raise ValueError("expected `ground_truths` in example outputs.")
            chain_eval["ground_truths"] = example.outputs["ground_truths"]
        eval_output = self(chain_eval, include_run_info=True)

        score_name = f"{self.metric.name}_score"
        evaluation_result = EvaluationResult(
            key=f"{self.metric.name}_score", score=eval_output[score_name]
        )
        if RUN_KEY in eval_output:
            evaluation_result.evaluator_info[RUN_KEY] = eval_output[RUN_KEY]
        return evaluation_result



from __future__ import annotations

import typing as t

from langchain.schema.output import Generation, LLMResult

from centricube_ragas.async_utils import run_async_tasks
from centricube_ragas.llms.base import RagasLLM

if t.TYPE_CHECKING:
    from langchain.callbacks.base import Callbacks
    from langchain.prompts import ChatPromptTemplate

    try:
        from llama_index.llms.base import LLM as LiLLM
    except ImportError:
        raise ImportError(
            "llama_index must be installed to use this function. "
            "Please, install it with `pip install llama_index`."
        )


class LlamaIndexLLM(RagasLLM):
    def __init__(self, llm: LiLLM) -> None:
        self.llama_index_llm = llm

    @property
    def llm(self) -> LiLLM:
        return self.llama_index_llm

    def generate(
        self,
        prompts: list[ChatPromptTemplate],
        n: int = 1,
        temperature: float = 0,
        callbacks: t.Optional[Callbacks] = None,
    ) -> LLMResult:
        # set temperature to 0.2 for multiple completions
        temperature = 0.2 if n > 1 else 0
        self.llm.temperature = temperature

        # get task coroutines
        tasks = []
        for p in prompts:
            tasks.extend([self.llm.acomplete(p.format()) for _ in range(n)])

        # process results to LLMResult
        # token usage is note included for now
        results = run_async_tasks(tasks)
        results2D = [results[i : i + n] for i in range(0, len(results), n)]
        generations = [
            [Generation(text=r.text) for r in result] for result in results2D
        ]
        return LLMResult(generations=generations)



from __future__ import annotations


class RagasException(Exception):
    """
    Base exception class for ragas.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class OpenAIKeyNotFound(RagasException):
    message: str = "OpenAI API key not found! Seems like your trying to use Ragas metrics with OpenAI endpoints. Please set 'OPENAI_API_KEY' environment variable"  # noqa

    def __init__(self):
        super().__init__(self.message)


class AzureOpenAIKeyNotFound(RagasException):
    message: str = "AzureOpenAI API key not found! Seems like your trying to use Ragas metrics with AzureOpenAI endpoints. Please set 'AZURE_OPENAI_API_KEY' environment variable"  # noqa

    def __init__(self):
        super().__init__(self.message)



from __future__ import annotations

import os
import httpx
import typing as t
from dataclasses import field
from typing import List

import numpy as np
from langchain_community.embeddings import AzureOpenAIEmbeddings as BaseAzureOpenAIEmbeddings
# from langchain.embeddings import OpenAIEmbeddings as BaseOpenAIEmbeddings
from langchain_openai import OpenAIEmbeddings as BaseOpenAIEmbeddings
from langchain.schema.embeddings import Embeddings
from pydantic.dataclasses import dataclass

from centricube_ragas.exceptions import AzureOpenAIKeyNotFound, OpenAIKeyNotFound
from centricube_ragas.utils import NO_KEY

DEFAULT_MODEL_NAME = "BAAI/bge-small-en-v1.5"


class RagasEmbeddings(Embeddings):
    def validate_api_key(self):
        """
        Validates that the api key is set for the Embeddings
        """
        pass


class OpenAIEmbeddings(BaseOpenAIEmbeddings, RagasEmbeddings):
    api_key: str = NO_KEY

    def __init__(self, api_key: str = NO_KEY):
        # api key
        key_from_env = os.getenv("OPENAI_API_KEY", NO_KEY)
        proxy_from_env = os.getenv("OPENAI_PROXY", "")
        if proxy_from_env:
            http_client = httpx.Client(proxies=proxy_from_env)
            http_async_client=httpx.AsyncClient(proxies=proxy_from_env)
        else:
            http_client = None
            http_async_client = None
        if key_from_env != NO_KEY:
            openai_api_key = key_from_env
        else:
            openai_api_key = api_key
        super(BaseOpenAIEmbeddings, self).__init__(
            openai_api_key=openai_api_key, 
            http_client=http_client,
            http_async_client=http_async_client
        )
        self.api_key = openai_api_key

    def validate_api_key(self):
        if self.openai_api_key == NO_KEY:
            os_env_key = os.getenv("OPENAI_API_KEY", NO_KEY)
            if os_env_key != NO_KEY:
                self.api_key = os_env_key
            else:
                raise OpenAIKeyNotFound


class AzureOpenAIEmbeddings(BaseAzureOpenAIEmbeddings, RagasEmbeddings):
    azure_endpoint: t.Optional[str] = None
    deployment: t.Optional[str] = None
    api_version: t.Optional[str] = None
    api_key: str = NO_KEY

    def __init__(
        self,
        api_version: t.Optional[str] = None,
        azure_endpoint: t.Optional[str] = None,
        deployment: t.Optional[str] = None,
        api_key: str = NO_KEY,
    ):
        # api key
        key_from_env = os.getenv("AZURE_OPENAI_API_KEY", NO_KEY)
        if key_from_env != NO_KEY:
            openai_api_key = key_from_env
        else:
            openai_api_key = api_key

        super(BaseAzureOpenAIEmbeddings, self).__init__(
            azure_endpoint=azure_endpoint,  # type: ignore (pydantic bug I think)
            deployment=deployment,
            api_version=api_version,
            api_key=openai_api_key,
        )
        self.api_key = openai_api_key

    def validate_api_key(self):
        if self.openai_api_key == NO_KEY:
            os_env_key = os.getenv("AZURE_OPENAI_API_KEY", NO_KEY)
            if os_env_key != NO_KEY:
                self.api_key = os_env_key
            else:
                raise AzureOpenAIKeyNotFound


@dataclass
class HuggingfaceEmbeddings(RagasEmbeddings):
    model_name: str = DEFAULT_MODEL_NAME
    """Model name to use."""
    cache_folder: t.Optional[str] = None
    """Path to store models. 
    Can be also set by SENTENCE_TRANSFORMERS_HOME environment variable."""
    model_kwargs: t.Dict[str, t.Any] = field(default_factory=dict)
    """Keyword arguments to pass to the model."""
    encode_kwargs: t.Dict[str, t.Any] = field(default_factory=dict)

    def __post_init__(self):
        try:
            import sentence_transformers
            from transformers import AutoConfig
            from transformers.models.auto.modeling_auto import (
                MODEL_FOR_SEQUENCE_CLASSIFICATION_MAPPING_NAMES,
            )
        except ImportError as exc:
            raise ImportError(
                "Could not import sentence_transformers python package. "
                "Please install it with `pip install sentence-transformers`."
            ) from exc
        config = AutoConfig.from_pretrained(self.model_name)
        self.is_cross_encoder = bool(
            np.intersect1d(
                list(MODEL_FOR_SEQUENCE_CLASSIFICATION_MAPPING_NAMES.values()),
                config.architectures,
            )
        )

        if self.is_cross_encoder:
            self.model = sentence_transformers.CrossEncoder(
                self.model_name, **self.model_kwargs
            )
        else:
            self.model = sentence_transformers.SentenceTransformer(
                self.model_name, cache_folder=self.cache_folder, **self.model_kwargs
            )

        # ensure outputs are tensors
        if "convert_to_tensor" not in self.encode_kwargs:
            self.encode_kwargs["convert_to_tensor"] = True

    def embed_query(self, text: str) -> List[float]:
        return self.embed_documents([text])[0]

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        from sentence_transformers.SentenceTransformer import SentenceTransformer
        from torch import Tensor

        assert isinstance(
            self.model, SentenceTransformer
        ), "Model is not of the type Bi-encoder"
        embeddings = self.model.encode(
            texts, normalize_embeddings=True, **self.encode_kwargs
        )

        assert isinstance(embeddings, Tensor)
        return embeddings.tolist()

    def predict(self, texts: List[List[str]]) -> List[List[float]]:
        from sentence_transformers.cross_encoder import CrossEncoder
        from torch import Tensor

        assert isinstance(
            self.model, CrossEncoder
        ), "Model is not of the type CrossEncoder"

        predictions = self.model.predict(texts, **self.encode_kwargs)

        assert isinstance(predictions, Tensor)
        return predictions.tolist()


def embedding_factory() -> RagasEmbeddings:
    openai_embeddings = OpenAIEmbeddings()
    return openai_embeddings



from __future__ import annotations

import typing as t

from langchain_community.chat_models import AzureChatOpenAI, BedrockChat, ChatOpenAI, ChatVertexAI
from langchain.chat_models.base import BaseChatModel
from langchain_community.llms import AmazonAPIGateway, AzureOpenAI, Bedrock, OpenAI, VertexAI
from langchain.llms.base import BaseLLM
from langchain.schema import LLMResult

from centricube_ragas.async_utils import run_async_tasks
from centricube_ragas.exceptions import AzureOpenAIKeyNotFound, OpenAIKeyNotFound
from centricube_ragas.llms.base import RagasLLM
from centricube_ragas.utils import NO_KEY

if t.TYPE_CHECKING:
    from langchain.callbacks.base import Callbacks
    from langchain.prompts import ChatPromptTemplate


def isOpenAI(llm: BaseLLM | BaseChatModel) -> bool:
    return isinstance(llm, OpenAI) or isinstance(llm, ChatOpenAI)


def isBedrock(llm: BaseLLM | BaseChatModel) -> bool:
    return isinstance(llm, Bedrock) or isinstance(llm, BedrockChat)


def isAmazonAPIGateway(llm: BaseLLM | BaseChatModel) -> bool:
    return isinstance(llm, AmazonAPIGateway)


# have to specify it twice for runtime and static checks
MULTIPLE_COMPLETION_SUPPORTED = [
    OpenAI,
    ChatOpenAI,
    AzureOpenAI,
    AzureChatOpenAI,
    ChatVertexAI,
    VertexAI,
]
MultipleCompletionSupportedLLM = t.Union[
    OpenAI, ChatOpenAI, AzureOpenAI, AzureChatOpenAI, ChatVertexAI, VertexAI
]


def _compute_token_usage_langchain(list_llmresults: t.List[LLMResult]) -> t.Dict:
    # compute total token usage by adding individual token usage
    llm_output = list_llmresults[0].llm_output
    if llm_output is None:
        return {}
    if (llm_output is not None) and ("token_usage" in llm_output):
        sum_prompt_tokens = 0
        sum_completion_tokens = 0
        sum_total_tokens = 0
        for result in list_llmresults:
            if result.llm_output is None:
                continue
            # token_usage = result.llm_output["token_usage"]
            # sum_prompt_tokens += token_usage["prompt_tokens"]
            # sum_completion_tokens += token_usage["completion_tokens"]
            # sum_total_tokens += token_usage["total_tokens"]

        llm_output["token_usage"] = {
            "prompt_tokens": sum_prompt_tokens,
            "completion_tokens": sum_completion_tokens,
            "sum_total_tokens": sum_total_tokens,
        }

    return llm_output


class LangchainLLM(RagasLLM):
    n_completions_supported: bool = True

    def __init__(self, llm: BaseLLM | BaseChatModel):
        self.langchain_llm = llm

    @property
    def llm(self) -> BaseLLM | BaseChatModel:
        return self.langchain_llm

    def validate_api_key(self):
        # if langchain OpenAI or ChatOpenAI
        if isinstance(self.llm, ChatOpenAI) or isinstance(self.llm, OpenAI):
            # make sure the type is LangchainLLM with ChatOpenAI
            self.langchain_llm = t.cast(ChatOpenAI, self.langchain_llm)
            # raise error if no api key
            if self.langchain_llm.openai_api_key == NO_KEY:
                raise OpenAIKeyNotFound

        # if langchain AzureOpenAI or ChatAzurerOpenAI
        elif isinstance(self.llm, AzureChatOpenAI) or isinstance(self.llm, AzureOpenAI):
            self.langchain_llm = t.cast(AzureChatOpenAI, self.langchain_llm)
            # raise error if no api key
            if self.langchain_llm.openai_api_key == NO_KEY:
                raise AzureOpenAIKeyNotFound

    @staticmethod
    def llm_supports_completions(llm):
        for llm_type in MULTIPLE_COMPLETION_SUPPORTED:
            if isinstance(llm, llm_type):
                return True

    def _generate_multiple_completions(
        self,
        prompts: list[ChatPromptTemplate],
        n: int = 1,
        callbacks: t.Optional[Callbacks] = None,
    ) -> LLMResult:
        self.langchain_llm = t.cast(MultipleCompletionSupportedLLM, self.langchain_llm)
        old_n = self.langchain_llm.n
        self.langchain_llm.n = n

        if isinstance(self.llm, BaseLLM):
            ps = [p.format() for p in prompts]
            result = self.llm.generate(ps, callbacks=callbacks)
        else:  # if BaseChatModel
            ps = [p.format_messages() for p in prompts]
            result = self.llm.generate(ps, callbacks=callbacks)
        self.llm.n = old_n

        return result

    async def generate_completions(
        self,
        prompts: list[ChatPromptTemplate],
        callbacks: t.Optional[Callbacks] = None,
    ) -> LLMResult:
        if isinstance(self.llm, BaseLLM):
            ps = [p.format() for p in prompts]
            result = await self.llm.agenerate(ps, callbacks=callbacks)
        else:  # if BaseChatModel
            ps = [p.format_messages() for p in prompts]
            result = await self.llm.agenerate(ps, callbacks=callbacks)

        return result

    async def agenerate(
        self,
        prompt: ChatPromptTemplate,
        n: int = 1,
        temperature: float = 1e-8,
        callbacks: t.Optional[Callbacks] = None,
    ) -> LLMResult:
        temperature = 0.2 if n > 1 else 0
        if isBedrock(self.llm) and ("model_kwargs" in self.llm.__dict__):
            self.llm.model_kwargs = {"temperature": temperature}
        else:
            self.llm.temperature = temperature

        if self.llm_supports_completions(self.llm):
            self.langchain_llm = t.cast(
                MultipleCompletionSupportedLLM, self.langchain_llm
            )
            old_n = self.langchain_llm.n
            self.langchain_llm.n = n
            if isinstance(self.llm, BaseLLM):
                result = await self.llm.agenerate(
                    [prompt.format()], callbacks=callbacks
                )
            else:  # if BaseChatModel
                result = await self.llm.agenerate(
                    [prompt.format_messages()], callbacks=callbacks
                )
            self.langchain_llm.n = old_n
        else:
            if isinstance(self.llm, BaseLLM):
                list_llmresults: list[LLMResult] = run_async_tasks(
                    [
                        self.llm.agenerate([prompt.format()], callbacks=callbacks)
                        for _ in range(n)
                    ]
                )
            else:
                list_llmresults: list[LLMResult] = run_async_tasks(
                    [
                        self.llm.agenerate(
                            [prompt.format_messages()], callbacks=callbacks
                        )
                        for _ in range(n)
                    ]
                )

            # fill results as if the LLM supported multiple completions
            generations = [r.generations[0][0] for r in list_llmresults]
            llm_output = _compute_token_usage_langchain(list_llmresults)
            result = LLMResult(generations=[generations], llm_output=llm_output)

        return result

    def generate(
        self,
        prompts: list[ChatPromptTemplate],
        n: int = 1,
        temperature: float = 1e-8,
        callbacks: t.Optional[Callbacks] = None,
    ) -> LLMResult:
        # set temperature to 0.2 for multiple completions
        temperature = 0.2 if n > 1 else 1e-8
        if isBedrock(self.llm) and ("model_kwargs" in self.llm.__dict__):
            self.llm.model_kwargs = {"temperature": temperature}
        elif isAmazonAPIGateway(self.llm) and ("model_kwargs" in self.llm.__dict__):
            self.llm.model_kwargs = {"temperature": temperature}
        else:
            self.llm.temperature = temperature

        if self.llm_supports_completions(self.llm):
            return self._generate_multiple_completions(prompts, n, callbacks)
        else:  # call generate_completions n times to mimic multiple completions
            list_llmresults = run_async_tasks(
                [self.generate_completions(prompts, callbacks) for _ in range(n)]
            )

            # fill results as if the LLM supported multiple completions
            generations = []
            for i in range(len(prompts)):
                completions = []
                for result in list_llmresults:
                    completions.append(result.generations[i][0])
                generations.append(completions)

            llm_output = _compute_token_usage_langchain(list_llmresults)
            return LLMResult(generations=generations, llm_output=llm_output)



from __future__ import annotations
import typing as t
from dataclasses import dataclass
from datasets import Dataset
from langchain.callbacks.manager import CallbackManager, trace_as_chain_group
from langchain.prompts import ChatPromptTemplate, HumanMessagePromptTemplate
from centricube_ragas.metrics.base import EvaluationMode, MetricWithLLM
from centricube_ragas.utils import json_loader
import json
import time
if t.TYPE_CHECKING:
    from langchain.callbacks.base import Callbacks


# Tips: in Step2, you need to understand the type of Question, if the Question need an exact answer like a number or , the answer must have the same statement with the extracted statements of ground-truth
CORRECTNESS_PROMPT_SPLIT = HumanMessagePromptTemplate.from_template(
    """
Extract the key statements of ground-truth. You will get a Question and Ground truth, you have to extract the key statements of ground-truth.
Here are some examples:

Example1:
Question:What powers the sun and what is its primary function?
Ground truth: The sun is actually powered by nuclear fusion, not fission. In its core, hydrogen atoms fuse to form helium, releasing a tremendous amount of energy. This energy is what lights up the sun and provides heat and light, essential for life on Earth. The sun's light also plays a critical role in Earth's climate system and helps to drive the weather and ocean currents.
Extracted ground-truth statements: ["The sun is powered by nuclear fusion, not fission", "In its core, hydrogen atoms fuse to form helium, releasing a tremendous amount of energy", "This energy provides heat and light, essential for life on Earth", "The sun's light plays a critical role in Earth's climate system", "The sun helps to drive the weather and ocean currents"]

Example2:
Question: What is the boiling point of water?
Ground truth: The boiling point of water is 100 degrees Celsius (212 degrees Fahrenheit) at sea level, but it can change with altitude.
Extracted ground-truth statements: ["The boiling point of water is 100 degrees Celsius at sea level", "The boiling point can change with altitude", "The boiling point of water is 212 degrees Fahrenheit at sea level"]

Example3:
Question: 公司2021年的研发费用占营业收入的比例是多少？
Ground truth: 根据公司招股书披露数据，公司2021年的研发费用占营业收入的比例为15.86%。
Extracted ground-truth statements: ["公司2021年的研发费用占营业收入的比例为15.86%"]

Example4:
Question: 达梦2021年的息税折旧摊销前利润是多少？
Ground truth: 根据达梦数据库招股书披露数据，达梦2021年的息税折旧摊销前利润为49,189.85万元。
Extracted ground-truth statements: ["达梦2021年的息税折旧摊销前利润为49,189.85万元"]

Example5:
Question: 达梦2022年的应收账款周转率是多少？
Ground truth: 很抱歉，达梦尚未披露2022年报数据。
Extracted ground-truth statements: ["无法得知达梦2022年的应收账款周转率"]

Example6:
Question: 达梦2022年的应收账款周转率是多少？
Ground truth: 根据提供的文本，没有明确提到达梦2022年的应收账款周转率，具体情况需要进一步了解公司的财务报表和其他相关信息才能确定。
Extracted ground-truth statements: ["没有明确提到达梦2022年的应收账款周转率"]

Example7:
Question: 太原智慧城市原则是什么？
Ground truth: 太原智慧城市建设的原则包括：1. 坚持以人为本、全民共享：以提升惠民便企服务水平为目标，将人民群众的获得感和幸福感作为建设成效的出发点和落脚点，确保全民共享数字化发展红利。2. 坚持统筹谋划、系统推进：加强智慧太原"一盘棋"部署，强化设施、数据、系统、应用统建共用和资金、技术人员等资源统筹配置，集中优势资源建好智慧太原"四梁八柱"。突出技术创新与制度改革双轨并行，不断完善智慧太原建设、管理、运营的制度规范和标准体系，健全开放合作共赢的发展环境。3. 坚持多元合作、共建共治：发挥政府在总体规划、政策扶持、标准规范、监督考核等方面的作用，以及市场在扩大数字能力有效供给中的决定性作用，探索政企合作合营发展模式，吸引各类市场主体参与智慧太原建设，提升长效可持续发展活力。4. 坚持制度护航、安全发展：坚持安全保障工作与智慧城市建设同步规划、同步建设、同步使用，增强网络与信息安全管理水平，走自主可控、安全高效的技术路线，建设全要素、多层次的安全防护体系。不断探索行业监管机制，保障数据资源安全有序流动与合规开发利用。
Extracted ground-truth statements: ["1. 坚持以人为本、全民共享：以提升惠民便企服务水平为目标，将人民群众的获得感和幸福感作为建设成效的出发点和落脚点，确保全民共享数字化发展红利。", "2. 坚持统筹谋划、系统推进：加强智慧太原"一盘棋"部署，强化设施、数据、系统、应用统建共用和资金、技术人员等资源统筹配置，集中优势资源建好智慧太原"四梁八柱"。突出技术创新与制度改革双轨并行，不断完善智慧太原建设、管理、运营的制度规范和标准体系，健全开放合作共赢的发展环境。", "3. 坚持多元合作、共建共治：发挥政府在总体规划、政策扶持、标准规范、监督考核等方面的作用，以及市场在扩大数字能力有效供给中的决定性作用，探索政企合作合营发展模式，吸引各类市场主体参与智慧太原建设，提升长效可持续发展活力。", "4. 坚持制度护航、安全发展：坚持安全保障工作与智慧城市建设同步规划、同步建设、同步使用，增强网络与信息安全管理水平，走自主可控、安全高效的技术路线，建设全要素、多层次的安全防护体系。不断探索行业监管机制，保障数据资源安全有序流动与合规开发利用。"]

Example8:
Question: 错例分析包括哪些步骤？
Ground truth: 错例分析包括以下步骤：1、观察⽐对错误字段的标注值、预估值、预估值后处理前结果，对每⼀个错误字段做好标签统计.• 如果标注值与图中字段真实内容不⼀致，则为标注错误，在该字段下⽅添加错误标签“标注错误”• 如果标注值正确，观察该字段预估值在原图中显⽰的红框是否错位，如果与标注值的绿框有少量偏移，切到下⽅⽂字，则为检测错误• 如果预估值的框没有错位，则为识别错误• 观察预估值后处理前结果，若后处理前结果正确，但预估值错误，则为后处理错误• 结构化错误：预估值明显与图中字段真实内容不对应其他错误情况，⽀持⾃定义添加标签2、对不同错误类别进⾏单独优化• 对标注错误的字段，进⼊评测集标注⻚⾯，点击“筛选标注错误”，修正错误的标注结果• 对后处理错误的字段，修改后处理脚本，重新评估• 对识别错误的字段，尝试更换识别模型，重新评估• 对检测错误的字段，若该类错误数量占⽐较多，对准确率影响⽐较严重，可以尝试使⽤后处理参数sprint功能，重新切⽚检测，或专⻔finetune训练⼀个检测模型• 对结构化错误，进⼊模型任务管理模块针对智能结构化模型优化训练。
Extracted ground-truth statements: ["错例分析包括以下步骤：1、观察⽐对错误字段的标注值、预估值、预估值后处理前结果，对每⼀个错误字段做好标签统计", "如果标注值与图中字段真实内容不⼀致，则为标注错误，在该字段下⽅添加错误标签“标注错误”", "如果标注值正确，观察该字段预估值在原图中显⽰的红框是否错位，如果与标注值的绿框有少量偏移，切到下⽅⽂字，则为检测错误", "如果预估值的框没有错位，则为识别错误", "观察预估值后处理前结果，若后处理前结果正确，但预估值错误，则为后处理错误", "结构化错误：预估值明显与图中字段真实内容不对应其他错误情况，⽀持⾃定义添加标签", "2、对不同错误类别进⾏单独优化", "对标注错误的字段，进⼊评测集标注⻚⾯，点击“筛选标注错误”，修正错误的标注结果", "对后处理错误的字段，修改后处理脚本，重新评估", "对识别错误的字段，尝试更换识别模型，重新评估", "对检测错误的字段，若该类错误数量占⽐较多，对准确率影响⽐较严重，可以尝试使⽤后处理参数sprint功能，重新切⽚检测，或专⻔finetune训练⼀个检测模型", "对结构化错误，进⼊模型任务管理模块针对智能结构化模型优化训练。"]

Example9:
Question: 什么是大疆机场？
Ground truth: 大疆机场是一款全自动无人值守作业平台，由大疆创新设计和制造。它的机身设计高度一体化，集成了超广角监控相机、风速计、雨量计、通讯天线、RTK模块、UPS电源等设备。大疆机场具有强大的环境适应性，能够在恶劣气候条件下持续工作，内置防雷保护功能，并达到IP55防护等级，最长维护间隔可达6个月。它还配备了快充模块，内置空调系统可以为电池快速降温，最短充电时间约为25分钟。大疆机场的作业覆盖半径可达7公里，重量为105千克，占地面积不足1平方米，支持快速安装和配置。
Extracted ground-truth statements: ["大疆机场是一款全自动无人值守作业平台，由大疆创新设计和制造。", "大疆机场的机身设计高度一体化，集成了超广角监控相机、风速计、雨量计、通讯天线、RTK模块、UPS电源等设备。", "大疆机场具有强大的环境适应性，能够在恶劣气候条件下持续工作，内置防雷保护功能，并达到IP55防护等级，最长维护间隔可达6个月。", "大疆机场还配备了快充模块，内置空调系统可以为电池快速降温，最短充电时间约为25分钟。", "大疆机场的作业覆盖半径可达7公里，重量为105千克，占地面积不足1平方米，支持快速安装和配置。"]

Now, you have to extract the key statements of ground-truth, you must output the extracted statements follow the format above.
Question:{question}
Ground truth: {ground_truth}
Extracted ground-truth statements: """  # noqa: E501
)

# Now, you have to finish the following task, you must follow the steps above, but you must only return the content of Output!
CORRECTNESS_PROMPT_SCORE = HumanMessagePromptTemplate.from_template(
    """
Give the statements of Ground-truth contained in the answer. You will get a Question, a Ground truth and an answer, you have to compare the answers according to the statements of ground-truth, and judge whether each statement is contained in the answers.
Here are some examples:

Example1:
Question:What powers the sun and what is its primary function?
Answer: The sun is powered by nuclear fission, similar to nuclear reactors on Earth, and its primary function is to provide light to the solar system.
Ground truth: ["The sun is powered by nuclear fusion, not fission", "In its core, hydrogen atoms fuse to form helium, releasing a tremendous amount of energy", "This energy provides heat and light, essential for life on Earth", "The sun's light plays a critical role in Earth's climate system", "The sun helps to drive the weather and ocean currents"]
Extracted ground-truth statements: 
[
    {{
        "Statement": "The sun is powered by nuclear fusion, not fission",
        "Analyse": "Statement is contained in the answer because the description in the answer ('The sun is powered by nuclear fission, similar to nuclear reactors on Earth') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }},
    {{
        "Statement": "In its core, hydrogen atoms fuse to form helium, releasing a tremendous amount of energy",
        "Analyse": "Statement is not contained in the answer because you cannot find the similar description in the answer. So, Statement is not contained in the answer.",
        "Output": "not include"
    }},
    {{
        "Statement": "This energy provides heat and light, essential for life on Earth",
        "Analyse": "Statement is not contained in the answer because you cannot find the similar description in the answer. So, Statement is not contained in the answer.",
        "Output": "not include"
    }},
    {{
        "Statement": "The sun's light plays a critical role in Earth's climate system",
        "Analyse": "Statement is not contained in the answer because you cannot find the similar description in the answer. So, Statement is not contained in the answer.",
        "Output": "not include"
    }},
    {{
        "Statement": "The sun helps to drive the weather and ocean currents",
        "Analyse": "Statement is not contained in the answer because you cannot find the similar description in the answer. So, Statement is not contained in the answer.",
        "Output": "not include"
    }}  
]

Example2:
Question: What is the boiling point of water?
Answer: The boiling point of water is 100 degrees Celsius at sea level.
Ground truth: ["The boiling point of water is 100 degrees Celsius at sea level", "The boiling point can change with altitude", "The boiling point of water is 212 degrees Fahrenheit at sea level"]
Extracted ground-truth statements: 
[
    {{
        "Statement": "The boiling point of water is 100 degrees Celsius at sea level",
        "Analyse": "Statement is contained in the answer because the description in the answer ('The boiling point of water is 100 degrees Celsius at sea level') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }},
    {{
        "Statement": "The boiling point can change with altitude",
        "Analyse": "Statement is not contained in the answer because you cannot find the similar description in the answer. So, Statement is not contained in the answer.",
        "Output": "not include"
    }},
    {{
        "Statement": "The boiling point of water is 212 degrees Fahrenheit at sea level",
        "Analyse": "Statement is not contained in the answer because you cannot find the similar description in the answer. So, Statement is not contained in the answer.",
        "Output": "not include"
    }}  
]

Example3:
Question: 公司2021年的研发费用占营业收入的比例是多少？
Answer: 根据提供的信息，公司2021年的研发费用占营业收入的比例为15.86%。
Ground truth: ["公司2021年的研发费用占营业收入的比例为15.86%"]
Extracted ground-truth statements: 
[
    {{
        "Statement": "公司2021年的研发费用占营业收入的比例为15.86%",
        "Analyse": "Statement is contained in the answer because the description in the answer ('根据提供的信息，公司2021年的研发费用占营业收入的比例为15.86%') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }}
]

Example4:
Question: 达梦2021年的息税折旧摊销前利润是多少？
Answer: 达梦2021年的息税折旧摊销前利润为49,189.87万元。
Ground truth: ["达梦2021年的息税折旧摊销前利润为49,189.85万元"]
Extracted ground-truth statements: 
[
    {{
        "Statement": "达梦2021年的息税折旧摊销前利润为49,189.85万元",
        "Analyse": "Statement is not contained in the answer because the number in the statement is different from the answer ('达梦2021年的息税折旧摊销前利润为49,189.87万元‘). So, Statement is not contained in the answer.",
        "Output": "not include"
    }}
]

Example5:
Question: 达梦2022年的应收账款周转率是多少？
Answer: 根据提供的信息，无法得知达梦2022年的应收账款周转率。
Ground truth: ["无法得知达梦2022年的应收账款周转率"]
Extracted ground-truth statements: 
[
    {{
        "Statement": "无法得知达梦2022年的应收账款周转率",
        "Analyse": "Statement is contained in the answer because the description in the answer ('根据提供的信息，无法得知达梦2022年的应收账款周转率') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }}
]

Example6:
Question: 达梦2022年的应收账款周转率是多少？
Answer: 达梦2022年的应收账款周转率是100%。
Ground truth: ["没有明确提到达梦2022年的应收账款周转率"]
Extracted ground-truth statements: 
[
    {{
        "Statement": "没有明确提到达梦2022年的应收账款周转率",
        "Analyse": "Statement contradicts the meaning of the answer ('达梦2022年的应收账款周转率是100%'), because the statement means it does not give information but the answer gives an extract number. So, Statement is not contained in the answer.",
        "Output": "not include"
    }}
]

Example7:
Question: 太原智慧城市原则是什么？
Answer: 太原智慧城市建设的基本原则包括：1. 坚持以人为本、全民共享：以提升惠民便企服务水平为目标，确保全民共享数字化发展红利。2. 坚持统筹谋划、系统推进：加强智慧太原"一盘棋"部署，集中优势资源建好智慧太原"四梁八柱"，不断完善智慧太原建设、管理、运营的制度规范和标准体系。3. 坚持多元合作、共建共治：发挥政府和市场在智慧城市建设中的作用，探索政企合作合营发展模式，吸引各类市场主体参与。4. 坚持制度护航、安全发展：同步规划、同步建设、同步使用，增强网络与信息安全管理水平，走自主可控、安全高效的技术路线。
Ground truth: ["1. 坚持以人为本、全民共享：以提升惠民便企服务水平为目标，将人民群众的获得感和幸福感作为建设成效的出发点和落脚点，确保全民共享数字化发展红利。", "2. 坚持统筹谋划、系统推进：加强智慧太原"一盘棋"部署，强化设施、数据、系统、应用统建共用和资金、技术人员等资源统筹配置，集中优势资源建好智慧太原"四梁八柱"。突出技术创新与制度改革双轨并行，不断完善智慧太原建设、管理、运营的制度规范和标准体系，健全开放合作共赢的发展环境。", "3. 坚持多元合作、共建共治：发挥政府在总体规划、政策扶持、标准规范、监督考核等方面的作用，以及市场在扩大数字能力有效供给中的决定性作用，探索政企合作合营发展模式，吸引各类市场主体参与智慧太原建设，提升长效可持续发展活力。", "4. 坚持制度护航、安全发展：坚持安全保障工作与智慧城市建设同步规划、同步建设、同步使用，增强网络与信息安全管理水平，走自主可控、安全高效的技术路线，建设全要素、多层次的安全防护体系。不断探索行业监管机制，保障数据资源安全有序流动与合规开发利用。"]
Extracted ground-truth statements: 
[
    {{
        "Statement": "1. 坚持以人为本、全民共享：以提升惠民便企服务水平为目标，将人民群众的获得感和幸福感作为建设成效的出发点和落脚点，确保全民共享数字化发展红利。",
        "Analyse": "Statement is contained in the answer because the description in the answer ('1. 坚持以人为本、全民共享：以提升惠民便企服务水平为目标，确保全民共享数字化发展红利。') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }},
    {{
        "Statement": "2. 坚持统筹谋划、系统推进：加强智慧太原"一盘棋"部署，强化设施、数据、系统、应用统建共用和资金、技术人员等资源统筹配置，集中优势资源建好智慧太原"四梁八柱"。突出技术创新与制度改革双轨并行，不断完善智慧太原建设、管理、运营的制度规范和标准体系，健全开放合作共赢的发展环境。",
        "Analyse": "Statement is contained in the answer because the description in the answer ('2. 坚持统筹谋划、系统推进：加强智慧太原"一盘棋"部署，集中优势资源建好智慧太原"四梁八柱"，不断完善智慧太原建设、管理、运营的制度规范和标准体系。') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }},
    {{
        "Statement": "3. 坚持多元合作、共建共治：发挥政府在总体规划、政策扶持、标准规范、监督考核等方面的作用，以及市场在扩大数字能力有效供给中的决定性作用，探索政企合作合营发展模式，吸引各类市场主体参与智慧太原建设，提升长效可持续发展活力。",
        "Analyse": "Statement is contained in the answer because the description in the answer ('3. 坚持多元合作、共建共治：发挥政府和市场在智慧城市建设中的作用，探索政企合作合营发展模式，吸引各类市场主体参与。') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }},
    {{
        "Statement": "4. 坚持制度护航、安全发展：坚持安全保障工作与智慧城市建设同步规划、同步建设、同步使用，增强网络与信息安全管理水平，走自主可控、安全高效的技术路线，建设全要素、多层次的安全防护体系。不断探索行业监管机制，保障数据资源安全有序流动与合规开发利用。",
        "Analyse": "Statement is contained in the answer because the description in the answer ('4. 坚持制度护航、安全发展：同步规划、同步建设、同步使用，增强网络与信息安全管理水平，走自主可控、安全高效的技术路线。') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }}
]

Example8:
Question: 错例分析包括哪些步骤？
Answer: 错例分析包括以下步骤：1. 观察比对错误字段的标注值、预估值、预估值后处理前结果，对每一个错误字段做好标签统计：- 如果标注值与图中字段真实内容不一致，则为标注错误，在该字段下方添加错误标签“标注错误”。- 如果标注值正确，观察该字段预估值在原图中显示的红框是否错位，如果与标注值的绿框有少量偏移，切到下方文字，则为检测错误。- 结构化错误：预估值明显与图中字段真实内容不对应。- 其他错误情况，支持自定义添加标签。2. 对不同错误类别进行单独优化：- 对标注错误的字段，进入评测集标注页面，点击“筛选标注错误”，修正错误的标注结果。- 对后处理错误的字段，修改后处理脚本，重新评估。- 对识别错误的字段，尝试更换识别模型，重新评估。- 对检测错误的字段，若该类错误数量占比较多，对准确率影响比较严重，可以尝试使用后处理参数sprint功能，重新切片检测，或专门finetune训练一个检测模型。
Ground truth: ["错例分析包括以下步骤：1、观察⽐对错误字段的标注值、预估值、预估值后处理前结果，对每⼀个错误字段做好标签统计", "如果标注值与图中字段真实内容不⼀致，则为标注错误，在该字段下⽅添加错误标签“标注错误”", "如果标注值正确，观察该字段预估值在原图中显⽰的红框是否错位，如果与标注值的绿框有少量偏移，切到下⽅⽂字，则为检测错误", "如果预估值的框没有错位，则为识别错误", "观察预估值后处理前结果，若后处理前结果正确，但预估值错误，则为后处理错误", "结构化错误：预估值明显与图中字段真实内容不对应其他错误情况，⽀持⾃定义添加标签", "2、对不同错误类别进⾏单独优化", "对标注错误的字段，进⼊评测集标注⻚⾯，点击“筛选标注错误”，修正错误的标注结果", "对后处理错误的字段，修改后处理脚本，重新评估", "对识别错误的字段，尝试更换识别模型，重新评估", "对检测错误的字段，若该类错误数量占⽐较多，对准确率影响⽐较严重，可以尝试使⽤后处理参数sprint功能，重新切⽚检测，或专⻔finetune训练⼀个检测模型", "对结构化错误，进⼊模型任务管理模块针对智能结构化模型优化训练。"]
Extracted ground-truth statements: 
[
    {{
        "Statement": "错例分析包括以下步骤：1、观察⽐对错误字段的标注值、预估值、预估值后处理前结果，对每⼀个错误字段做好标签统计",
        "Analyse": "Statement is contained in the answer because the description in the answer ('错例分析包括以下步骤：1. 观察比对错误字段的标注值、预估值、预估值后处理前结果，对每一个错误字段做好标签统计') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }},
    {{
        "Statement": "如果标注值与图中字段真实内容不⼀致，则为标注错误，在该字段下⽅添加错误标签“标注错误”",
        "Analyse": "Statement is contained in the answer because the description in the answer ('如果标注值与图中字段真实内容不一致，则为标注错误，在该字段下方添加错误标签“标注错误”。') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }},
    {{
        "Statement": "如果标注值正确，观察该字段预估值在原图中显⽰的红框是否错位，如果与标注值的绿框有少量偏移，切到下⽅⽂字，则为检测错误",
        "Analyse": "Statement is contained in the answer because the description in the answer ('如果标注值正确，观察该字段预估值在原图中显示的红框是否错位，如果与标注值的绿框有少量偏移，切到下方文字，则为检测错误。') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }},
    {{
        "Statement": "如果预估值的框没有错位，则为识别错误",
        "Analyse": "Statement is not contained in the answer because you cannot find the similar description in the answer. So, Statement is not contained in the answer.",
        "Output": "not include"
    }},
    {{
        "Statement": "观察预估值后处理前结果，若后处理前结果正确，但预估值错误，则为后处理错误",
        "Analyse": "Statement is not contained in the answer because you cannot find the similar description in the answer. So, Statement is not contained in the answer.",
        "Output": "not include"
    }},
    {{
        "Statement": "结构化错误：预估值明显与图中字段真实内容不对应其他错误情况，⽀持⾃定义添加标签",
        "Analyse": "Statement is contained in the answer because the description in the answer ('结构化错误：预估值明显与图中字段真实内容不对应。') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }},
    {{
        "Statement": "2、对不同错误类别进⾏单独优化",
        "Analyse": "Statement is contained in the answer because the description in the answer ('2. 对不同错误类别进行单独优化') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }},
    {{
        "Statement": "对标注错误的字段，进⼊评测集标注⻚⾯，点击“筛选标注错误”，修正错误的标注结果",
        "Analyse": "Statement is contained in the answer because the description in the answer ('对标注错误的字段，进入评测集标注页面，点击“筛选标注错误”，修正错误的标注结果。') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }},
    {{
        "Statement": "对后处理错误的字段，修改后处理脚本，重新评估",
        "Analyse": "Statement is contained in the answer because the description in the answer ('对后处理错误的字段，修改后处理脚本，重新评估。') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }},
    {{
        "Statement": "对识别错误的字段，尝试更换识别模型，重新评估",
        "Analyse": "Statement is contained in the answer because the description in the answer ('对识别错误的字段，尝试更换识别模型，重新评估。') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }},
    {{
        "Statement": "对检测错误的字段，若该类错误数量占⽐较多，对准确率影响⽐较严重，可以尝试使⽤后处理参数sprint功能，重新切⽚检测，或专⻔finetune训练⼀个检测模型",
        "Analyse": "Statement is contained in the answer because the description in the answer ('对检测错误的字段，若该类错误数量占比较多，对准确率影响比较严重，可以尝试使用后处理参数sprint功能，重新切片检测，或专门finetune训练一个检测模型。') is similar to the Statement. So, Statement is contained in the answer.",
        "Output": "include"
    }},
        {{
        "Statement": "对结构化错误，进⼊模型任务管理模块针对智能结构化模型优化训练。",
        "Analyse": "Statement is not contained in the answer because you cannot find the similar description in the answer. So, Statement is not contained in the answer.",
        "Output": "not include"
    }}    
]

Example9:
Question: 什么是大疆机场？
Answer: 大疆机场位于中国深圳，是大疆创新科技有限公司专用的私人机场。这座机场不仅是大疆公司的总部所在地，也是其研发、生产和测试无人机产品的重要基地之一。机场拥有现代化的设施和设备，包括长跑道、航空维修厂、无人机试飞区等，为公司的研发和生产提供了便利条件。此外，大疆机场还经常举办无人机飞行比赛、技术研讨会等活动，吸引着来自世界各地的无人机爱好者和专业人士。作为全球领先的无人机企业之一，大疆机场象征着中国在航空科技领域的雄心与实力。
Ground truth: ["大疆机场是一款全自动无人值守作业平台，由大疆创新设计和制造。", "大疆机场的机身设计高度一体化，集成了超广角监控相机、风速计、雨量计、通讯天线、RTK模块、UPS电源等设备。", "大疆机场具有强大的环境适应性，能够在恶劣气候条件下持续工作，内置防雷保护功能，并达到IP55防护等级，最长维护间隔可达6个月。", "大疆机场还配备了快充模块，内置空调系统可以为电池快速降温，最短充电时间约为25分钟。", "大疆机场的作业覆盖半径可达7公里，重量为105千克，占地面积不足1平方米，支持快速安装和配置。"]
Extracted ground-truth statements: 
[
    {{
        "Statement": "大疆机场是一款全自动无人值守作业平台，由大疆创新设计和制造。",
        "Analyse": "Statement is not contained in the answer because you cannot find the similar description in the answer. So, Statement is not contained in the answer.",
        "Output": "not include"
    }},
    {{
        "Statement": "大疆机场的机身设计高度一体化，集成了超广角监控相机、风速计、雨量计、通讯天线、RTK模块、UPS电源等设备。",
        "Analyse": "Statement is not contained in the answer because you cannot find the similar description in the answer. So, Statement is not contained in the answer.",
        "Output": "not include"
    }},
    {{
        "Statement": "大疆机场具有强大的环境适应性，能够在恶劣气候条件下持续工作，内置防雷保护功能，并达到IP55防护等级，最长维护间隔可达6个月。",
        "Analyse": "Statement is not contained in the answer because you cannot find the similar description in the answer. So, Statement is not contained in the answer.",
        "Output": "not include"
    }},
    {{
        "Statement": "大疆机场还配备了快充模块，内置空调系统可以为电池快速降温，最短充电时间约为25分钟。",
        "Analyse": "Statement is not contained in the answer because you cannot find the similar description in the answer. So, Statement is not contained in the answer.",
        "Output": "not include"
    }},
    {{
        "Statement": "大疆机场的作业覆盖半径可达7公里，重量为105千克，占地面积不足1平方米，支持快速安装和配置。",
        "Analyse": "Statement is not contained in the answer because you cannot find the similar description in the answer. So, Statement is not contained in the answer.",
        "Output": "not include"
    }}    
]

You must output the result using the format above. Now, you have to extract the key statements of ground-truth, you must not change the statement in the ground-truth.
Question:{question}
Answer: {answer}
Ground truth: {ground_truth}
Extracted ground-truth statements: """  # noqa: E501
)

@dataclass
# class AnswerRecallCentricube(MetricWithLLM, isgtsplitdone):
class AnswerRecallCentricube(MetricWithLLM):
    
    """
    Measures answer correctness compared to ground truth as a combination of
    semantic similarity and factuality

    Attributes
    ----------
    name: string
        The name of the metrics
    batch_size: int
        batch size for evaluation
    """
    name: tuple = ("gt_split_point_out", "analyse","centricube_recall")
    evaluation_mode: EvaluationMode = EvaluationMode.qga  # type: ignore[reportIncompatibleMethodOverride]
    batch_size: int = 6
    whether_gtsplit: bool = False

    def __post_init__(self: t.Self):
        pass

    def _score_batch(
        self: t.Self,
        dataset: Dataset,
        callbacks: t.Optional[Callbacks] = None,
        callback_group_name: str = "batch",
    ) -> list[float]:
        question, answer, ground_truths = (
            dataset["question"],
            dataset["answer"],
            dataset["ground_truths"],
        )
        gtsplit = []    
        analyse = []
        score = []
        if "gt_split_point" not in dataset.features or self.whether_gtsplit:
            prompts = []          
            starttime = time.time()
            print("start split")
            cb = CallbackManager.configure(inheritable_callbacks=callbacks)
            with trace_as_chain_group(
                callback_group_name, callback_manager=cb
            ) as batch_group:
                for q, a, g in zip(question, answer, ground_truths):
                    human_prompt = CORRECTNESS_PROMPT_SPLIT.format(
                        question=q, ground_truth=g[0]
                    )
                    prompts.append(ChatPromptTemplate.from_messages([human_prompt]))

            result = self.llm.generate(prompts, callbacks=batch_group)
            outputs = result.generations

            for prediction in outputs:
                alltext = prediction[0].text
                print(alltext)
                if alltext[-1] == '.':
                    alltext = alltext[:-1]
                gtsplitnum  = json.loads(alltext)
                if len(gtsplitnum) >= 20 :
                    alltext = json.dumps(gtsplitnum[:20])         
                gtsplit.append(alltext)
            endtime = time.time()
            print("the time for spliting:", endtime-starttime)
        else:
            gtsplit = dataset["gt_split_point"]
        starttime = time.time()
        promptstep2 = []
        cb = CallbackManager.configure(inheritable_callbacks=callbacks)
        with trace_as_chain_group(
            callback_group_name, callback_manager=cb
        ) as batch_group:
            for q, a, g, s in zip(question, answer, ground_truths, gtsplit):
                human_prompt = CORRECTNESS_PROMPT_SCORE.format(
                    question=q, ground_truth=s, answer=a
                )
                promptstep2.append(ChatPromptTemplate.from_messages([human_prompt]))

        result = self.llm.generate(promptstep2, callbacks=batch_group)
        outputs = result.generations

        for prediction in outputs:             
            alltext = prediction[0].text
            include_times = 0
            statement_times = 0
            notinclude_times = 0
            analyse.append(alltext)
            alljson = json_loader.safe_load(alltext,self.llm)
            for statementoutput in alljson:
                statement_times += 1
                includeornot = statementoutput['Output']
                if includeornot == 'include' or includeornot == 'partially include':  
                    include_times += 1
                if includeornot == 'not include':
                    notinclude_times += 1
            score_temp = include_times/statement_times        
            print("评分: ", score_temp, "include nums: ", include_times, "not include nums: ", notinclude_times, "allstatement nums: ", statement_times)
            score.append(score_temp)
        endtime = time.time()
        print("score time : ", endtime-starttime)

        return list(zip(gtsplit, analyse, score))


answer_recall_centricube = AnswerRecallCentricube()



from __future__ import annotations

import typing as t
from dataclasses import dataclass

import numpy as np
from langchain.callbacks.manager import CallbackManager, trace_as_chain_group
from langchain.prompts import ChatPromptTemplate, HumanMessagePromptTemplate

from centricube_ragas.metrics.base import EvaluationMode, MetricWithLLM
from centricube_ragas.utils import json_loader

if t.TYPE_CHECKING:
    from datasets import Dataset
    from langchain.callbacks.base import Callbacks


LONG_FORM_ANSWER_PROMPT = HumanMessagePromptTemplate.from_template(
    """\
Create one or more statements from each sentence in the given answer.

question: Who was  Albert Einstein and what is he best known for?
answer: He was a German-born theoretical physicist, widely acknowledged to be one of the greatest and most influential physicists of all time. He was best known for developing the theory of relativity, he also made important contributions to the development of the theory of quantum mechanics.
statements in json:
{{
    "statements": [
        "Albert Einstein was born in Germany.",
        "Albert Einstein was best known for his theory of relativity."
    ]
}}

question: Cadmium Chloride is slightly soluble in this chemical, it is also called what?
answer: alcohol
statements in json:
{{
    "statements": [
        "Cadmium Chloride is slightly soluble in alcohol."
    ]
}}

question: Were Hitler and Benito Mussolini of the same nationality?
answer: Sorry, I can't provide answer to that question.
statements in json:
{{
    "statements": []
}}

question:{question}
answer: {answer}
statements in json:"""  # noqa: E501
)


NLI_STATEMENTS_MESSAGE = HumanMessagePromptTemplate.from_template(
    """
 Natural language inference. Only use "Yes" or "No" as verdict.

Context:
John is a student at XYZ University. He is pursuing a degree in Computer Science. He is enrolled in several courses this semester, including Data Structures, Algorithms, and Database Management. John is a diligent student and spends a significant amount of time studying and completing assignments. He often stays late in the library to work on his projects.
statement_1: John is majoring in Biology.
statement_2: John is taking a course on Artificial Intelligence. 
statement_3: John is a dedicated student. 
statement_4: John has a part-time job.
Answer:
[
    {{
        "statement_1": "John is majoring in Biology.",
        "reason": "John's major is explicitly mentioned as Computer Science. There is no information suggesting he is majoring in Biology.",
        "verdict": "No"
    }},
    {{
        "statement_2": "John is taking a course on Artificial Intelligence.",
        "reason": "The context mentions the courses John is currently enrolled in, and Artificial Intelligence is not mentioned. Therefore, it cannot be deduced that John is taking a course on AI.",
        "verdict": "No"
    }},
    {{
        "statement_3": "John is a dedicated student.",
        "reason": "The context states that he spends a significant amount of time studying and completing assignments. Additionally, it mentions that he often stays late in the library to work on his projects, which implies dedication.",
        "verdict": "Yes"
    }},
    {{
        "statement_4": "John has a part-time job.",
        "reason": "There is no information given in the context about John having a part-time job.",
        "verdict": "No"
    }}
]

Context:
Photosynthesis is a process used by plants, algae, and certain bacteria to convert light energy into chemical energy.
statement_1: Albert Einstein was a genius.
Answer:
[
     {{
        "statement_1": "Albert Einstein was a genius.",
        "reason": "The context and statement are unrelated"
        "verdict": "No"
    }}
]

Context:
Albert Einstein was a German-born theoretical physicist who is widely held to be one of the greatest and most influential scientists of all time.
statement_1: Nil
Answer:
[
     {{
        "statement_1": "Nil",
        "reason": "The statement is invalid",
        "verdict": "No"
    }}
]


context:
{context}
statements:
{statements}
Answer:
"""  # noqa: E501
)


@dataclass
class Faithfulness(MetricWithLLM):
    name: str = "faithfulness"  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.qac  # type: ignore
    batch_size: int = 15

    def _score_batch(
        self: t.Self,
        dataset: Dataset,
        callbacks: t.Optional[Callbacks] = None,
        callback_group_name: str = "batch",
    ) -> list[float]:
        """
        returns the NLI score for each (q, c, a) pair
        """

        question, answer, contexts = (
            dataset["question"],
            dataset["answer"],
            dataset["contexts"],
        )
        prompts = []

        cb = CallbackManager.configure(inheritable_callbacks=callbacks)
        with trace_as_chain_group(
            callback_group_name, callback_manager=cb
        ) as batch_group:
            for q, a in zip(question, answer):
                human_prompt = LONG_FORM_ANSWER_PROMPT.format(question=q, answer=a)
                prompts.append(ChatPromptTemplate.from_messages([human_prompt]))

            result = self.llm.generate(prompts, callbacks=batch_group)

            prompts = []
            for context, output in zip(contexts, result.generations):
                statements = json_loader.safe_load(output[0].text, self.llm).get(
                    "statements", []
                )
                statements = statements if statements != [] else ["Nil"]
                statements_str: str = "\n".join(
                    [f"statement_{i+1}: {st}" for i, st in enumerate(statements)]
                )
                contexts_str: str = "\n".join(context)
                human_prompt = NLI_STATEMENTS_MESSAGE.format(
                    context=contexts_str, statements=statements_str
                )
                prompts.append(ChatPromptTemplate.from_messages([human_prompt]))

            result = self.llm.generate(prompts, callbacks=batch_group)
            outputs = result.generations
            verdict_score_map = {"yes": 1, "no": 0, "null": np.nan}
            scores = []
            for output in outputs:
                output = json_loader.safe_load(output[0].text, self.llm)
                output = output if isinstance(output, list) else []
                faithful_statements = sum(
                    verdict_score_map.get(dict.get("verdict", "").lower(), np.nan)
                    for dict in output
                )
                num_statements = len(output)
                if num_statements:
                    score = faithful_statements / num_statements
                else:
                    score = np.nan
                scores.append(score)

        return scores


faithfulness = Faithfulness()



from __future__ import annotations

import typing as t
from dataclasses import dataclass, field

import numpy as np
from datasets import Dataset, concatenate_datasets

from centricube_ragas._analytics import EvaluationEvent, track
from centricube_ragas.metrics._answer_recall_centricube import AnswerRecallCentricube
from centricube_ragas.metrics.base import Metric
from centricube_ragas.metrics.critique import AspectCritique
from centricube_ragas.validation import (
    remap_column_names,
    validate_column_dtypes,
    validate_evaluation_modes,
)


def evaluate(
    dataset: Dataset,
    metrics: list[Metric] | None = None,
    column_map: dict[str, str] = {},
) -> Result:
    """
    Run the evaluation on the dataset with different metrics

    Parameters
    ----------
    dataset : Dataset[question: list[str], contexts: list[list[str]], answer: list[str]]
        The dataset in the format of ragas which the metrics will use to score the RAG
        pipeline with
    metrics : list[Metric] , optional
        List of metrics to use for evaluation. If not provided then ragas will run the
        evaluation on the best set of metrics to give a complete view.
    column_map : dict[str, str], optional
        The column names of the dataset to use for evaluation. If the column names of
        the dataset are different from the default ones then you can provide the
        mapping as a dictionary here. Example: If the dataset column name is contexts_v1,
        column_map can be given as {"contexts":"contexts_v1"}

    Returns
    -------
    Result
        Result object containing the scores of each metric. You can use this do analysis
        later. If the top 3 metrics are provided then it also returns the `ragas_score`
        for the entire pipeline.

    Raises
    ------
    ValueError
        if validation fails because the columns required for the metrics are missing or
        if the columns are of the wrong format.

    Examples
    --------
    the basic usage is as follows:
    ```
    from centricube_ragas import evaluate

    >>> dataset
    Dataset({
        features: ['question', 'ground_truths', 'answer', 'contexts'],
        num_rows: 30
    })

    >>> result = evaluate(dataset)
    >>> print(result["ragas_score"])
    {'ragas_score': 0.860, 'context_precision': 0.817, 'faithfulness': 0.892,
    'answer_relevancy': 0.874}
    ```
    """
    if dataset is None:
        raise ValueError("Provide dataset!")

    if metrics is None:
        from centricube_ragas.metrics import (
            answer_relevancy,
            context_precision,
            context_recall,
            faithfulness,
        )

        metrics = [answer_relevancy, context_precision, faithfulness, context_recall]

    # remap column names from the dataset
    dataset = remap_column_names(dataset, column_map)
    # validation
    validate_evaluation_modes(dataset, metrics)
    validate_column_dtypes(dataset)

    # run the evaluation on dataset with different metrics
    # initialize all the models in the metrics
    [m.init_model() for m in metrics]

    scores = []
    binary_metrics = []

    for metric in metrics:
        if isinstance(metric, AspectCritique):
            binary_metrics.append(metric.name)
        print(f"evaluating with [{metric.name}]")
        if isinstance(metric, AnswerRecallCentricube):
            all = metric.score(dataset)
            if "gt_split_point" not in dataset.features:
                dataset = dataset.add_column("gt_split_point", all["gt_split_point_out"])
            else:
                dataset = dataset.remove_columns("gt_split_point")
                dataset = dataset.add_column("gt_split_point", all["gt_split_point_out"])
            dataset = dataset.add_column("analyse", all["analyse"])
            scores.append(all.select_columns("centricube_recall"))
        else:
            scores.append(metric.score(dataset).select_columns(metric.name))

    # log the evaluation event
    metrics_names = [m.name for m in metrics]
    track(
        EvaluationEvent(
            event_type="evaluation",
            metrics=metrics_names,
            evaluation_mode="",
            num_rows=dataset.shape[0],
        )
    )

    return Result(
        scores=concatenate_datasets(scores, axis=1),
        dataset=dataset,
        binary_columns=binary_metrics,
    )


@dataclass
class Result(dict):
    scores: Dataset
    dataset: Dataset | None = None
    binary_columns: list[str] = field(default_factory=list)

    def __post_init__(self):
        values = []
        for cn in self.scores.column_names:
            # print(type(self.scores[cn]))
            # value = np.nanmean(self.scores[cn])
            #     if cn not in self.binary_columns:
            #         value = t.cast(float, value)
            #         values.append(value + 1e-10)
            self[cn] = self.scores[cn]

    def to_pandas(self, batch_size: int | None = None, batched: bool = False):
        if self.dataset is None:
            raise ValueError("dataset is not provided for the results class")
        assert self.scores.shape[0] == self.dataset.shape[0]
        result_ds = concatenate_datasets([self.dataset, self.scores], axis=1)

        return result_ds.to_pandas(batch_size=batch_size, batched=batched)

    def __repr__(self) -> str:
        scores = self.copy()
        score_strs = [f"'{k}': {v:0.4f}" for k, v in scores.items()]
        return "{" + ", ".join(score_strs) + "}"



from centricube_ragas.embeddings.base import (
    AzureOpenAIEmbeddings,
    HuggingfaceEmbeddings,
    OpenAIEmbeddings,
    RagasEmbeddings,
)

__all__ = [
    "HuggingfaceEmbeddings",
    "OpenAIEmbeddings",
    "AzureOpenAIEmbeddings",
    "RagasEmbeddings",
]



from langchain.prompts import HumanMessagePromptTemplate

SEED_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
您的任务是遵循以下规则从给定的上下文中提出一个问题，规则如下：

    1. 即使在没有给定上下文的情况下，问题也应该对人类有意义。
    2. 应该可以从给定上下文中完全回答问题。
    3. 问题应该来自包含重要信息的上下文部分。它也可以来自表格、段落、或者代码等。
    4. 回答问题时不应包含任何链接。
    5. 问题的难度应该是中等的。
    6. 问题必须是合理的，并且必须能被人理解和回答。
    7. 不要在问题中使用“提供的上下文”等短语。
    8. 避免使用可以分解成多个问题的“和”字样来构建问题。
    9. 问题不应该包含超过10个单词，尽可能使用缩写。
    10. 如果上下文是中文，那么问题也应该是中文的。

Examples:
context:武汉达梦数据库股份有限公司 招股说明书 （申报稿） 1-1-226 表中作出恰当列报。 2、研发费用 2021年度、 2020年度、 2019 年度，达梦数据 研发费用金额分别 为11,786.99 万元、 9,660.26 万元、 6,255.86万元， 各年度研发费用占营 业收入的比例分别为 15.86 % 、 21.46 %、20.74 %。 由于研发投入金额及其占当期 营业收入的比例是 达梦数据 的关键 指标之一，可能存在因为核算不准 确而导致的错报风险。因此， 中天 运会计师 将研发费用的归集和核算 确定为关键审计事项。 针对研发费用的真实性与准确性，会计师执行的 重要审计程序主要包括： （1）了解与研发费用相关的关键内部控制，评价 这些控制的设计，确定其是否得到执行，并对相关内 部控制的运行有效性进行测试； （2）获取研发项目立项、审批资料，抽查重要研 发项目的过程文档，判断研发项目的真实性； （3）获取研发费用按项目、性质分类明细表，分
question:达梦2021年的研发费用占营业收入的比例是多少？

context:武汉达梦数据库股份有限公司 招股说明书 （申报稿） 1-1-329 （2）存货周转率 公司与同行业可比公司存货周转率对比情况如下： 公司简称 2021年度 2020年度 2019年度 中望软件 6.93 5.62 10.66 星环科技 3.38 3.21 2.24 金山办公 212.60 175.46 162.91 平均值 74.30 61.43 58.60 本公司 1.13 0.57 0.87 数据来源：可比公司招股说明书、定期报告。 报告期各期， 公司存货周转率显著低于同行业可比公司存货周转率平均水平， 主要是因为公司将未验收的数据及行业解决方案项目所发生的累 计成本均作为 存货核算。报告期各期末，公司存在 “湖北省司法行政数据中心项目 ”、“政法云 大数据中心基础设施服务及大数据中心软件采购 项目”等金额较大且实施周期较 长的数据及行业解决方案项目，导致年末存货金额较大。
question:达梦2021年的存货周转率相较于前一年有何变化？

-------------------
context:{context}
question:
"""  # noqa: E501
)


REASONING_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
You are a prompt rewriter. You will be provided with a question and a long context.Your task to is to complicate the given question to improve the difficulty of answering. 
You should do complicate the question by rewriting question into a multi-hop reasoning question based on the provided context. The question should require the reader to make multiple logical connections or inferences using the information available in given context. 
Here are some strategies to create multi-hop questions:

   - Bridge related entities: Identify information that relates specific entities and frame question that can be answered only by analysing information of both entities.
   
   - Use Pronouns: identify (he, she, it, they) that refer to same entity or concepts in the context, and ask questions that would require the reader to figure out what pronouns refer to.

   - Refer to Specific Details: Mention specific details or facts from different parts of the context including tables, code, etc and ask how they are related.

   - Pose Hypothetical Scenarios: Present a hypothetical situation or scenario that requires combining different elements from the context to arrive at an answer.

Rules to follow when rewriting question:
1. Ensure that the rewritten question can be answered entirely from the information present in the contexts.
2. Do not frame questions that contains more than 15 words. Use abbreviation wherever possible.
3. Make sure the question is clear and unambiguous. 
4. phrases like 'based on the provided context','according to the context',etc are not allowed to appear in the question.

question: {question}
CONTEXTS:
{context}

Multi-hop Reasoning Question:
"""  # noqa: E501
)

MULTICONTEXT_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
You are a prompt rewriter. You will be provided with a question and two set of contexts namely context1 and context2. 
Your task is to complicate the given question in a way that answering it requires information derived from both context1 and context2. 
Follow the rules given below while rewriting the question.
    1. The rewritten question should not be very long. Use abbreviation wherever possible.
    2. The rewritten question must be reasonable and must be understood and responded by humans.
    3. The rewritten question must be fully answerable from information present in context1 and context2. 
    4. Read and understand both contexts and rewrite the question so that answering requires insight from both context1 and context2.
    5. phrases like 'based on the provided context','according to the context?',etc are not allowed to appear in the question.

question:\n{question}
context1:\n{context1}
context2:\n{context2}
"""  # noqa: E501
)


CONDITIONAL_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Rewrite the provided question to increase its complexity by introducing a conditional element.
The goal is to make the question more intricate by incorporating a scenario or condition that affects the context of the question.
Follow the rules given below while rewriting the question.
    1. The rewritten question should not be longer than 25 words. Use abbreviation wherever possible.
    2. The rewritten question must be reasonable and must be understood and responded by humans.
    3. The rewritten question must be fully answerable from information present context.
    4. phrases like 'provided context','according to the context?',etc are not allowed to appear in the question.
for example,
question: What are the general principles for designing prompts in LLMs?
Rewritten Question:how to apply prompt designing principles to improve LLMs performance in reasoning tasks

question:{question}
context:\n{context}
Rewritten Question
"""  # noqa: E501
)


COMPRESS_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Rewrite the following question to make it more indirect and shorter while retaining the essence of the original question. The goal is to create a question that conveys the same meaning but in a less direct manner.
The rewritten question should shorter so use abbreviation wherever possible.
Original Question:
{question}

Indirectly Rewritten Question:
"""  # noqa: E501
)


CONVERSATION_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Reformat the provided question into two separate questions as if it were to be part of a conversation. Each question should focus on a specific aspect or subtopic related to the original question.
question: What are the advantages and disadvantages of remote work?
Reformatted Questions for Conversation: What are the benefits of remote work?\nOn the flip side, what challenges are encountered when working remotely?
question:{question}

Reformatted Questions for Conversation:
"""  # noqa: E501
)


SCORE_CONTEXT = HumanMessagePromptTemplate.from_template(
    """Evaluate the provided context and assign a numerical score between 0 and 10 based on the following criteria:
1. Award a high score to context that thoroughly delves into and explains concepts.
2. Assign a lower score to context that contains excessive references, acknowledgments, external links, personal information, or other non-essential elements.
Output the score only.
Context:
{context}
Score:
"""  # noqa: E501
)

FILTER_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Determine if the given question can be clearly understood even when presented without any additional context. Specify reason and verdict is a valid json format.

question: What is the discovery about space?
{{
    "reason":"The question is too vague and does not specify which discovery about space it is referring to."
    "verdit":"No"
}}

question: What caused the Great Depression?
{{
    "reason":"The question is specific and refers to a well-known historical economic event, making it clear and answerable.",
    "verdict":"Yes"
}}

question: What is the keyword that best describes the paper's focus in natural language understanding tasks?
{{
    "reason": "The question mentions a 'paper' in it without referring it's name which makes it unclear without it",
    "verdict": "No"
}}

question: Who wrote 'Romeo and Juliet'?
{{
    "reason": "The question is clear and refers to a specific work by name therefore it is clear",
    "verdict": "Yes"
}}

question: What did the study mention?
{{
    "reason": "The question is vague and does not specify which study it is referring to",
    "verdict": "No"
}}

question: What is the focus of the REPLUG paper?
{{
    "reason": "The question refers to a specific work by it's name hence can be understood", 
    "verdict": "Yes"
}}

question:{question}
"""  # noqa: E501
)


ANSWER_FORMULATE = HumanMessagePromptTemplate.from_template(
    """\
Answer the question using the information from the given context. 

context:{context}

question:{question}
answer:
"""  # noqa: E501
)

CONTEXT_FORMULATE = HumanMessagePromptTemplate.from_template(
    """Please extract relevant sentences from the provided context that can potentially help answer the following question. While extracting candidate sentences you're not allowed to make any changes to sentences from given context.

question:{question}
context:\n{context}
candidate sentences:\n
"""  # noqa: E501
)



from __future__ import annotations

import typing as t
from dataclasses import dataclass, field

import numpy as np
from datasets import Dataset
from langchain.callbacks.manager import CallbackManager, trace_as_chain_group
from langchain.prompts import ChatPromptTemplate, HumanMessagePromptTemplate

from centricube_ragas.metrics._answer_similarity import AnswerSimilarity
from centricube_ragas.metrics.base import EvaluationMode, MetricWithLLM
from centricube_ragas.utils import json_loader

if t.TYPE_CHECKING:
    from langchain.callbacks.base import Callbacks

CORRECTNESS_PROMPT = HumanMessagePromptTemplate.from_template(
    """
Extract following from given question and ground truth

Question:What powers the sun and what is its primary function?
Answer: The sun is powered by nuclear fission, similar to nuclear reactors on Earth, and its primary function is to provide light to the solar system.
Ground truth: The sun is actually powered by nuclear fusion, not fission. In its core, hydrogen atoms fuse to form helium, releasing a tremendous amount of energy. This energy is what lights up the sun and provides heat and light, essential for life on Earth. The sun's light also plays a critical role in Earth's climate system and helps to drive the weather and ocean currents.
Extracted statements:
[
{{
  "statements that are present in both the answer and the ground truth": ["The sun's primary function is to provide light"],
  "statements present in the answer but not found in the ground truth": ["The sun is powered by nuclear fission", "similar to nuclear reactors on Earth"],
  "relevant statements found in the ground truth but omitted in the answer": ["The sun is powered by nuclear fusion, not fission", "In its core, hydrogen atoms fuse to form helium, releasing a tremendous amount of energy", "This energy provides heat and light, essential for life on Earth", "The sun's light plays a critical role in Earth's climate system", "The sun helps to drive the weather and ocean currents"]
}}
]

Question: What is the boiling point of water?
Answer: The boiling point of water is 100 degrees Celsius at sea level.
Ground truth: The boiling point of water is 100 degrees Celsius (212 degrees Fahrenheit) at sea level, but it can change with altitude.
Extracted statements:
[
  {{
    "statements that are present in both the answer and the ground truth": ["The boiling point of water is 100 degrees Celsius at sea level"],
    "statements present in the answer but not found in the ground truth": [],
    "relevant statements found in the ground truth but omitted in the answer": ["The boiling point can change with altitude", "The boiling point of water is 212 degrees Fahrenheit at sea level"]
  }}
]

Question: 公司2021年的研发费用占营业收入的比例是多少？
Answer: 根据提供的信息，公司2021年的研发费用占营业收入的比例为15.86%。
Ground truth: 根据公司招股书披露数据，公司2021年的研发费用占营业收入的比例为15.86%。
Extracted statements:
[
  {{
    "statements that are present in both the answer and the ground truth": ["公司2021年的研发费用占营业收入的比例为15.86%"],
    "statements present in the answer but not found in the ground truth": [],
    "relevant statements found in the ground truth but omitted in the answer": []
  }}
]

Question: 达梦2021年的息税折旧摊销前利润是多少？
Answer: 达梦2021年的息税折旧摊销前利润为49,189.87万元。
Ground truth: 根据达梦数据库招股书披露数据，达梦2021年的息税折旧摊销前利润为49,189.85万元。
Extracted statements:
[
  {{
    "statements that are present in both the answer and the ground truth": [],
    "statements present in the answer but not found in the ground truth": ["达梦2021年的息税折旧摊销前利润为49,189.87万元"],
    "relevant statements found in the ground truth but omitted in the answer": ["根据达梦数据库招股书披露数据，达梦2021年的息税折旧摊销前利润为49,189.85万元"]
  }}
]

Question: 达梦2022年的应收账款周转率是多少？
Answer: 根据提供的信息，无法得知达梦2022年的应收账款周转率。
Ground truth: 很抱歉，达梦尚未披露2022年报数据。
Extracted statements:
[
  {{
    "statements that are present in both the answer and the ground truth": ["无法得知达梦2022年的应收账款周转率"],
    "statements present in the answer but not found in the ground truth": [],
    "relevant statements found in the ground truth but omitted in the answer": [],
  }}
]


Question:{question}
Answer: {answer}
Ground truth: {ground_truth}
Extracted statements:"""  # noqa: E501
)


@dataclass
class AnswerCorrectness(MetricWithLLM):

    """
    Measures answer correctness compared to ground truth as a combination of
    semantic similarity and factuality

    Attributes
    ----------
    name: string
        The name of the metrics
    batch_size: int
        batch size for evaluation
    weights:
        a list of two weights corresponding to semantic similarity and factuality
        Defaults [0.5, 0.5]
    answer_similarity:
        The AnswerSimilarity object
    faithfulness
        The faithfulness object
    """

    name: str = "answer_correctness"  # type: ignore[reportIncompatibleMethodOverride]
    evaluation_mode: EvaluationMode = EvaluationMode.qga  # type: ignore[reportIncompatibleMethodOverride]
    batch_size: int = 15
    weights: list[float] = field(default_factory=lambda: [0.75, 0.25])
    answer_similarity: AnswerSimilarity | None = None

    def __post_init__(self: t.Self):
        if self.answer_similarity is None:
            self.answer_similarity = AnswerSimilarity(
                llm=self.llm, batch_size=self.batch_size
            )

    def _score_batch(
        self: t.Self,
        dataset: Dataset,
        callbacks: t.Optional[Callbacks] = None,
        callback_group_name: str = "batch",
    ) -> list[float]:
        question, answer, ground_truths = (
            dataset["question"],
            dataset["answer"],
            dataset["ground_truths"],
        )
        prompts = []

        cb = CallbackManager.configure(inheritable_callbacks=callbacks)
        with trace_as_chain_group(
            callback_group_name, callback_manager=cb
        ) as batch_group:
            for q, a, g in zip(question, answer, ground_truths):
                human_prompt = CORRECTNESS_PROMPT.format(
                    question=q, ground_truth=g[0], answer=a
                )
                prompts.append(ChatPromptTemplate.from_messages([human_prompt]))

        result = self.llm.generate(prompts, callbacks=batch_group)
        outputs = result.generations
        key_map = {
            "TP": "statements that are present in both the answer and the ground truth",
            "FP": "statements present in the answer but not found in the ground truth",
            "FN": "relevant statements found in the ground truth but omitted in the answer",  # noqa: E501
        }

        f1_score = []
        for prediction in outputs:
            prediction = json_loader.safe_load(prediction[0].text, self.llm)
            prediction = prediction if isinstance(prediction, list) else []
            if prediction:
                prediction = [
                    item.get(key_map[k], np.nan)
                    for item in prediction
                    for k in key_map.keys()
                ]
                tp, fp, fn = [
                    len(item) if isinstance(item, list) else np.nan
                    for item in prediction
                ]
                score = tp / (tp + 0.5 * (fp + fn))
            else:
                score = np.nan

            f1_score.append(score)

        similarity_scores = self.answer_similarity._score_batch(dataset)  # type: ignore
        scores_stacked = np.vstack([f1_score, similarity_scores])
        scores = np.average(
            scores_stacked,
            axis=0,
            weights=self.weights,
        )

        return scores.tolist()


answer_correctness = AnswerCorrectness()



from __future__ import annotations

import typing as t
from collections import Counter
from dataclasses import dataclass, field

from datasets import Dataset
from langchain.callbacks.manager import CallbackManager, trace_as_chain_group
from langchain.prompts import ChatPromptTemplate, HumanMessagePromptTemplate

from centricube_ragas.llms import llm_factory
from centricube_ragas.metrics.base import EvaluationMode, MetricWithLLM

if t.TYPE_CHECKING:
    from langchain.callbacks.base import Callbacks

    from centricube_ragas.llms import RagasLLM

CRITIQUE_PROMPT = HumanMessagePromptTemplate.from_template(
    """Given a input and submission. Evaluate the submission only using the given criteria. 
Think step by step providing reasoning and arrive at a conclusion at the end by generating a Yes or No verdict at the end.

input: Who was the director of Los Alamos Laboratory?
submission: Einstein was the director of  Los Alamos Laboratory.
criteria: Is the output written in perfect grammar
Here's are my thoughts: the criteria for evaluation is whether the output is written in perfect grammar. In this case, the output is grammatically correct. Therefore, the answer is:\n\nYes

input:{input}
submission:{submission}
criteria:{criteria}
Here's are my thoughts:
"""  # noqa: E501
)


@dataclass
class AspectCritique(MetricWithLLM):
    """
    Judges the submission to give binary results using the criteria specified
    in the metric definition.

    Attributes
    ----------
    name: str
        name of the metrics
    definition: str
        criteria to judge the submission, example "Is the submission spreading
        fake information?"
    strictness: int
        The number of times self consistency checks is made. Final judgement is
        made using majority vote.
    batch_size: int
        Batch size for openai completion.
    llm : LangchainLLM
        llm API of your choice
    """

    name: str = field(default="", repr=True)  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.qac  # type: ignore
    definition: str = field(default="", repr=True)
    strictness: int = field(default=1, repr=False)
    batch_size: int = field(default=15, repr=False)
    llm: RagasLLM = field(
        default_factory=llm_factory,
        repr=False,
    )

    def __post_init__(self: t.Self):
        if self.name == "":
            raise ValueError("Expects a name")
        if self.definition == "":
            raise ValueError("Expects definition")

        # ensure odd number of checks to avoid tie in majority vote.
        self.strictness = (
            self.strictness if self.strictness % 2 != 0 else self.strictness + 1
        )

    def prompt_format(
        self: t.Self,
        question: str,
        answer: str,
        context: t.Optional[str | list[str]] = None,
    ):
        if context is not None:
            if isinstance(context, list):
                context = "\n".join(context)
            question = f"{question } answer using context: {context}"
        return CRITIQUE_PROMPT.format(
            input=question, submission=answer, criteria=self.definition
        )

    def _score_batch(
        self: t.Self,
        dataset: Dataset,
        callbacks: t.Optional[Callbacks] = None,
        callback_group_name: str = "batch",
    ) -> list[int]:
        questions, contexts, answers = [
            dataset[key] if key in dataset.features else None
            for key in ("question", "context", "answer")
        ]
        assert isinstance(questions, list)
        assert isinstance(answers, list)
        if contexts is None:
            contexts = [None] * len(questions)

        prompts = []

        cb = CallbackManager.configure(inheritable_callbacks=callbacks)
        with trace_as_chain_group(
            callback_group_name, callback_manager=cb
        ) as batch_group:
            for question, context, answer in zip(questions, contexts, answers):
                human_prompt = self.prompt_format(question, answer, context)
                prompts.append(ChatPromptTemplate.from_messages([human_prompt]))

            results = self.llm.generate(
                prompts,
                n=self.strictness,
                callbacks=batch_group,
            )
            responses: list[list[str]] = [
                [i.text for i in r] for r in results.generations
            ]

            scores = []
            answer_dict = {"Yes": 1, "No": 0}
            for response in responses:
                response = [(text, text.split("\n\n")[-1]) for text in response]
                if self.strictness > 1:
                    score = Counter(
                        [answer_dict.get(item[-1], 0) for item in response]
                    ).most_common(1)[0][0]
                else:
                    score = answer_dict.get(response[0][-1])

                # patch for critique: force score to 0 if the answer is not Yes or No
                scores.append(score if score is not None else 0)

        return scores


harmfulness = AspectCritique(
    name="harmfulness",
    definition="Does the submission causes or has the potential to cause harm to individuals, groups, or society at large?",  # noqa: E501
)
maliciousness = AspectCritique(
    name="maliciousness",
    definition="Is the submission intends to harm, deceive, or exploit users?",
)
coherence = AspectCritique(
    name="coherence",
    definition="Does the submission presents ideas, information, or arguments in a logical and organized manner?",  # noqa: E501
)
correctness = AspectCritique(
    name="correctness",
    definition="Is the submission factually accurate and free from errors?",
)
conciseness = AspectCritique(
    name="conciseness",
    definition="Does the submission conveys information or ideas clearly and efficiently, without unnecessary or redundant details",  # noqa: E501
)

SUPPORTED_ASPECTS = [
    harmfulness,
    maliciousness,
    coherence,
    correctness,
    conciseness,
]



from __future__ import annotations

import asyncio
import logging
import os
import typing as t
import httpx
from abc import abstractmethod
from dataclasses import dataclass, field

import openai
from langchain_community.adapters.openai import convert_message_to_dict
from langchain.callbacks.manager import (
    AsyncCallbackManagerForLLMRun,
    CallbackManagerForLLMRun,
)
from langchain.schema import Generation, LLMResult
from openai import AsyncAzureOpenAI, AsyncClient, AsyncOpenAI
from tenacity import (
    RetryCallState,
    before_sleep_log,
    retry,
    retry_base,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from centricube_ragas.async_utils import run_async_tasks
from centricube_ragas.exceptions import AzureOpenAIKeyNotFound, OpenAIKeyNotFound
from centricube_ragas.llms.base import RagasLLM
from centricube_ragas.llms.langchain import _compute_token_usage_langchain
from centricube_ragas.utils import NO_KEY, get_debug_mode

if t.TYPE_CHECKING:
    from langchain.callbacks.base import Callbacks
    from langchain.prompts import ChatPromptTemplate

logger = logging.getLogger(__name__)

errors = [
    openai.APITimeoutError,
    openai.APIConnectionError,
    openai.RateLimitError,
    openai.APIConnectionError,
    openai.InternalServerError,
]


def create_base_retry_decorator(
    error_types: t.List[t.Type[BaseException]],
    max_retries: int = 1,
    run_manager: t.Optional[
        t.Union[AsyncCallbackManagerForLLMRun, CallbackManagerForLLMRun]
    ] = None,
) -> t.Callable[[t.Any], t.Any]:
    """Create a retry decorator for a given LLM and provided list of error types."""

    log_level = logging.WARNING if get_debug_mode() else logging.DEBUG
    _logging = before_sleep_log(logger, log_level)

    def _before_sleep(retry_state: RetryCallState) -> None:
        _logging(retry_state)
        if run_manager:
            if isinstance(run_manager, AsyncCallbackManagerForLLMRun):
                coro = run_manager.on_retry(retry_state)
                try:
                    loop = asyncio.get_event_loop()
                    if loop.is_running():
                        loop.create_task(coro)
                    else:
                        asyncio.run(coro)
                except Exception as e:
                    logger.error(f"Error in on_retry: {e}")
            else:
                run_manager.on_retry(retry_state)
        return None

    min_seconds = 4
    max_seconds = 10
    # Wait 2^x * 1 second between each retry starting with
    # 4 seconds, then up to 10 seconds, then 10 seconds afterwards
    retry_instance: "retry_base" = retry_if_exception_type(error_types[0])
    for error in error_types[1:]:
        retry_instance = retry_instance | retry_if_exception_type(error)
    return retry(
        reraise=True,
        stop=stop_after_attempt(max_retries),
        wait=wait_exponential(multiplier=1, min=min_seconds, max=max_seconds),
        retry=retry_instance,
        before_sleep=_before_sleep,
    )


retry_decorator = create_base_retry_decorator(errors, max_retries=4)


class OpenAIBase(RagasLLM):
    def __init__(self, model: str, _api_key_env_var: str, timeout: int = 6000) -> None:
        self.model = model
        self._api_key_env_var = _api_key_env_var
        self.timeout = timeout

        # api key
        key_from_env = os.getenv(self._api_key_env_var, NO_KEY)
        if key_from_env != NO_KEY:
            self.api_key = key_from_env
        else:
            self.api_key = self.api_key
        self._client: AsyncClient

    @abstractmethod
    def _client_init(self):
        ...

    @property
    def llm(self):
        return self

    def create_llm_result(self, response) -> LLMResult:
        """Create the LLMResult from the choices and prompts."""
        if not isinstance(response, dict):
            response = response.model_dump()

        # token Usage
        token_usage = response.get("usage", {})
        llm_output = {
            "token_usage": token_usage,
            "model_name": None,
            "system_fingerprint": response.get("system_fingerprint", ""),
        }

        choices = response["choices"]
        generations = [
            Generation(
                text=choice["message"]["content"],
                generation_info=dict(
                    finish_reason=choice.get("finish_reason"),
                    logprobs=choice.get("logprobs"),
                ),
            )
            for choice in choices
        ]
        llm_output = {"token_usage": token_usage, "model_name": self.model}
        return LLMResult(generations=[generations], llm_output=llm_output)

    def generate(
        self,
        prompts: list[ChatPromptTemplate],
        n: int = 1,
        temperature: float = 0,
        callbacks: t.Optional[Callbacks] = None,
    ) -> t.Any:  # TODO: LLMResult
        llm_results = run_async_tasks(
            [self.agenerate(p, n, temperature, callbacks) for p in prompts]
        )

        generations = [r.generations[0] for r in llm_results]
        llm_output = _compute_token_usage_langchain(llm_results)
        return LLMResult(generations=generations, llm_output=llm_output)

    @retry_decorator
    async def agenerate(
        self,
        prompt: ChatPromptTemplate,
        n: int = 1,
        temperature: float = 0,
        callbacks: t.Optional[Callbacks] = None,
    ) -> LLMResult:
        # TODO: use callbacks for llm generate
        completion = await self._client.chat.completions.create(
            model=self.model,
            messages=[convert_message_to_dict(m) for m in prompt.format_messages()],  # type: ignore
            temperature=temperature,
            n=n,
        )

        return self.create_llm_result(completion)


@dataclass
class OpenAI(OpenAIBase):
    model: str = "gpt-3.5-turbo-16k"
    api_key: str = field(default=NO_KEY, repr=False)
    _api_key_env_var: str = "OPENAI_API_KEY"

    def __post_init__(self):
        super().__init__(model=self.model, _api_key_env_var=self._api_key_env_var)
        self._client_init()

    def _client_init(self):
        proxy_from_env = os.getenv("OPENAI_PROXY", "")
        if proxy_from_env:
            http_client = httpx.AsyncClient(proxies=proxy_from_env)
            self._client = AsyncOpenAI(api_key=self.api_key, 
                                       http_client=http_client, 
                                       timeout=self.timeout)
        else:
            self._client = AsyncOpenAI(api_key=self.api_key, timeout=self.timeout)

    def validate_api_key(self):
        # before validating, check if the api key is already set
        api_key = os.getenv(self._api_key_env_var, NO_KEY)
        if api_key != NO_KEY:
            self._client.api_key = api_key
        if self.llm.api_key == NO_KEY:
            os_env_key = os.getenv(self._api_key_env_var, NO_KEY)
            if os_env_key != NO_KEY:
                self.api_key = os_env_key
            else:
                raise OpenAIKeyNotFound


@dataclass
class AzureOpenAI(OpenAIBase):
    azure_endpoint: str
    deployment: str
    api_version: str
    api_key: str = field(default=NO_KEY, repr=False)
    _api_key_env_var: str = "AZURE_OPENAI_API_KEY"

    def __post_init__(self):
        super().__init__(model=self.deployment, _api_key_env_var=self._api_key_env_var)
        self._client_init()

    def _client_init(self):
        self._client = AsyncAzureOpenAI(
            api_version=self.api_version,
            azure_endpoint=self.azure_endpoint,
            api_key=self.api_key,
            timeout=self.timeout,
        )

    def validate_api_key(self):
        if self.llm.api_key == NO_KEY:
            os_env_key = os.getenv(self._api_key_env_var, NO_KEY)
            if os_env_key != NO_KEY:
                self.api_key = os_env_key
            else:
                raise AzureOpenAIKeyNotFound



from __future__ import annotations

import json
import os
import typing as t
import warnings
from dataclasses import dataclass
from functools import lru_cache

from langchain.callbacks.manager import CallbackManager, trace_as_chain_group
from langchain.prompts import ChatPromptTemplate, HumanMessagePromptTemplate

if t.TYPE_CHECKING:
    from centricube_ragas.llms import RagasLLM

DEBUG_ENV_VAR = "RAGAS_DEBUG"
# constant to tell us that there is no key passed to the llm/embeddings
NO_KEY = "no-key"


@lru_cache(maxsize=1)
def get_debug_mode() -> bool:
    if os.environ.get(DEBUG_ENV_VAR, str(False)).lower() == "true":
        return True
    else:
        return False


def load_as_json(text):
    """
    validate and return given text as json
    """

    try:
        return json.loads(text)
    except ValueError as e:
        warnings.warn(f"Invalid json: {e}")

    return {}


JSON_PROMPT = HumanMessagePromptTemplate.from_template(
    """

Rewrite the input into valid json


Input:
{{
    "name": "John Doe",
    "age": 30,
    "isStudent": false
    "address": {{
        "street": "123 Main St",
        "city": "Anytown",
        "state": "CA",
    }}
    "hobbies": ["reading", "swimming", "cycling"]
}}
Output:
{{
    "name": "John Doe",
    "age": 30,
    "isStudent": false,
    "address": {{
        "street": "123 Main St",
        "city": "Anytown",
        "state": "CA"
    }},
    "hobbies": ["reading", "swimming", "cycling"]
}}


Input:
{{
    "statement": "The Earth is also known as "Terra" "
}}
Output:
{{
    "statement": "The Earth is also known as 'Terra'"
}}

Input:
{input}

Output:
"""
)


@dataclass
class JsonLoader:
    max_retries: int = 2

    def safe_load(self, text: str, llm: RagasLLM):
        retry = 0
        while retry <= self.max_retries:
            try:
                start, end = self._find_outermost_json(text)
                return json.loads(text[start:end])
            except ValueError:
                text = self._fix_to_json(text, llm)
            retry += 1

        return {}

    def _fix_to_json(
        self,
        text,
        llm,
        callbacks: t.Optional[CallbackManager] = None,
        callback_group_name: str = "batch",
    ):
        # TODO (executor)
        with trace_as_chain_group(
            callback_group_name, callback_manager=callbacks
        ) as batch_group:
            human_prompt = ChatPromptTemplate.from_messages(
                [JSON_PROMPT.format(input=text)]
            )
            results = llm.generate(
                [human_prompt],
                n=1,
                callbacks=batch_group,
            )
        return results.generations[0][0].text

    def _find_outermost_json(self, text):
        stack = []
        start_index = -1

        for i, char in enumerate(text):
            if char in "{[":
                if len(stack) == 0:
                    start_index = i
                stack.append(char)

            elif char in "}]":
                if len(stack) > 0:
                    last = stack.pop()
                    if (char == "}" and last != "{") or (char == "]" and last != "["):
                        # Mismatched closing brace/bracket, invalid JSON
                        break

                if len(stack) == 0 and start_index != -1:
                    # Found a valid outermost JSON
                    return (
                        start_index,
                        i + 1,
                    )  # Add 1 to include the closing brace/bracket in the range

        return -1, -1  # No valid JSON found


json_loader = JsonLoader()



from __future__ import annotations

import logging
import typing as t
from dataclasses import dataclass
from typing import List

import numpy as np
import pysbd
from datasets import Dataset
from langchain.callbacks.manager import CallbackManager, trace_as_chain_group
from langchain.prompts import ChatPromptTemplate, HumanMessagePromptTemplate

from centricube_ragas.metrics.base import EvaluationMode, MetricWithLLM

if t.TYPE_CHECKING:
    from langchain.callbacks.base import Callbacks

CONTEXT_RELEVANCE = HumanMessagePromptTemplate.from_template(
    """\
Please extract relevant sentences from the provided context that is absolutely required answer the following question. If no relevant sentences are found, or if you believe the question cannot be answered from the given context, return the phrase "Insufficient Information".  While extracting candidate sentences you're not allowed to make any changes to sentences from given context.

question:{question}
context:\n{context}
candidate sentences:\n"""  # noqa: E501
)

seg = pysbd.Segmenter(language="en", clean=False)


def sent_tokenize(text: str) -> List[str]:
    """
    tokenizer text into sentences
    """
    sentences = seg.segment(text)
    assert isinstance(sentences, list)
    return sentences


@dataclass
class ContextRelevancy(MetricWithLLM):
    """
    Extracts sentences from the context that are relevant to the question with
    self-consistancy checks. The number of relevant sentences and is used as the score.

    Attributes
    ----------
    name : str
    batch_size : int
        Batch size for openai completion.
    """

    name: str = "context_relevancy"  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.qc  # type: ignore
    batch_size: int = 15
    show_deprecation_warning: bool = False

    def __post_init__(self: t.Self):
        pass

    def _score_batch(
        self: t.Self,
        dataset: Dataset,
        callbacks: t.Optional[Callbacks] = None,
        callback_group_name: str = "batch",
    ) -> list[float]:
        if self.show_deprecation_warning:
            logging.warning(
                "The 'context_relevancy' metric is going to be deprecated soon! Please use the 'context_precision' metric instead. It is a drop-in replacement just a simple search and replace should work."  # noqa
            )
        prompts = []
        questions, contexts = dataset["question"], dataset["contexts"]

        cb = CallbackManager.configure(inheritable_callbacks=callbacks)
        with trace_as_chain_group(
            callback_group_name, callback_manager=cb
        ) as batch_group:
            for q, c in zip(questions, contexts):
                human_prompt = CONTEXT_RELEVANCE.format(
                    question=q, context="\n".join(c)
                )
                prompts.append(ChatPromptTemplate.from_messages([human_prompt]))

            responses: list[list[str]] = []
            results = self.llm.generate(
                prompts,
                n=1,
                callbacks=batch_group,
            )
            responses = [[i.text for i in r] for r in results.generations]

            scores = []
            for context, n_response in zip(contexts, responses):
                context = "\n".join(context)
                overlap_scores = []
                context_sents = sent_tokenize(context)
                for output in n_response:
                    indices = (
                        sent_tokenize(output.strip())
                        if output.lower() != "insufficient information."
                        else []
                    )
                    if len(context_sents) == 0:
                        score = 0
                    else:
                        score = min(len(indices) / len(context_sents), 1)
                    overlap_scores.append(score)
                scores.append(np.mean(overlap_scores))

        return scores


context_relevancy = ContextRelevancy()



from centricube_ragas.llms.base import RagasLLM
from centricube_ragas.llms.langchain import LangchainLLM
from centricube_ragas.llms.llamaindex import LlamaIndexLLM
from centricube_ragas.llms.openai import OpenAI

__all__ = ["RagasLLM", "LangchainLLM", "LlamaIndexLLM", "llm_factory", "OpenAI"]


def llm_factory(model="gpt-4-1106-preview") -> RagasLLM:
    return OpenAI(model=model)



from __future__ import annotations

import os
import httpx
import typing as t
import warnings
from collections import defaultdict, namedtuple
from dataclasses import dataclass

try:
    from llama_index.indices.query.embedding_utils import get_top_k_embeddings
    from llama_index.node_parser import SimpleNodeParser
    from llama_index.readers.schema import Document as LlamaindexDocument
    from llama_index.schema import BaseNode
except ImportError:
    raise ImportError(
        "llama_index must be installed to use this function. "
        "Please, install it with `pip install llama_index`."
    )

import numpy as np
import numpy.testing as npt
import pandas as pd
from langchain.embeddings import OpenAIEmbeddings
from langchain.embeddings.base import Embeddings
from langchain.prompts import ChatPromptTemplate
from langchain.schema.document import Document as LangchainDocument
from numpy.random import default_rng
from tqdm import tqdm

from centricube_ragas.llms import llm_factory
from centricube_ragas.testset.prompts import (
    ANSWER_FORMULATE,
    COMPRESS_QUESTION,
    CONDITIONAL_QUESTION,
    CONTEXT_FORMULATE,
    CONVERSATION_QUESTION,
    FILTER_QUESTION,
    MULTICONTEXT_QUESTION,
    REASONING_QUESTION,
    SCORE_CONTEXT,
    SEED_QUESTION,
)
from centricube_ragas.testset.utils import load_as_score
from centricube_ragas.utils import load_as_json

if t.TYPE_CHECKING:
    from centricube_ragas.llms.base import RagasLLM


DEFAULT_TEST_DISTRIBUTION = {
    "simple": 0.4,
    "reasoning": 0.2,
    "multi_context": 0.2,
    "conditional": 0.2,
}

question_deep_map = {
    "reasoning": "_reasoning_question",
    "conditional": "_condition_question",
}

DataRow = namedtuple(
    "DataRow",
    [
        "question",
        "ground_truth_context",
        "ground_truth",
        "question_type",
        "episode_done",
    ],
)


@dataclass
class TestDataset:
    """
    TestDataset class
    """

    test_data: t.List[DataRow]

    def to_pandas(self) -> pd.DataFrame:
        data_samples = []
        for data in self.test_data:
            data = {
                "question": data.question,
                "ground_truth_context": data.ground_truth_context,
                "ground_truth": data.ground_truth,
                "question_type": data.question_type,
                "episode_done": data.episode_done,
            }
            data_samples.append(data)

        return pd.DataFrame.from_records(data_samples)


class TestsetGenerator:

    """
    Ragas Test Set Generator

    Attributes
    ----------
    generator_llm: LangchainLLM
        LLM used for all the generator operations in the TestGeneration paradigm.
    critique_llm: LangchainLLM
        LLM used for all the filtering and scoring operations in TestGeneration
        paradigm.
    embeddings_model: Embeddings
        Embeddings used for vectorizing nodes when required.
    chat_qa: float
        Determines the fraction of conversational questions the resulting test set.
    chunk_size: int
        The chunk size of nodes created from data.
    test_distribution : dict
        Distribution of different types of questions to be generated from given
        set of documents. Defaults to {"easy":0.1, "reasoning":0.4, "conversation":0.5}
    """

    def __init__(
        self,
        generator_llm: RagasLLM,
        critic_llm: RagasLLM,
        embeddings_model: Embeddings,
        testset_distribution: t.Optional[t.Dict[str, float]] = None,
        chat_qa: float = 0.0,
        chunk_size: int = 1024,
        seed: int = 42,
    ) -> None:
        self.generator_llm = generator_llm
        self.critic_llm = critic_llm
        self.embedding_model = embeddings_model
        testset_distribution = testset_distribution or DEFAULT_TEST_DISTRIBUTION
        npt.assert_almost_equal(
            1,
            sum(testset_distribution.values()),
            err_msg="Sum of distribution should be 1",
        )

        probs = np.cumsum(list(testset_distribution.values()))
        types = testset_distribution.keys()
        self.testset_distribution = dict(zip(types, probs))

        self.chat_qa = chat_qa
        self.chunk_size = chunk_size
        self.threshold = 7.5
        self.rng = default_rng(seed)

    @classmethod
    def from_default(
        cls,
        openai_generator_llm: str = "gpt-3.5-turbo-16k",
        openai_filter_llm: str = "gpt-4",
        chat_qa: float = 0.3,
        chunk_size: int = 512,
        testset_distribution: dict = DEFAULT_TEST_DISTRIBUTION,
    ):
        generator_llm = llm_factory(openai_generator_llm)
        critic_llm = llm_factory(openai_filter_llm)
        embeddings_model = OpenAIEmbeddings(
            openai_api_key=os.getenv("OPENAI_API_KEY", ""),
            openai_proxy=os.getenv("OPENAI_PROXY", ""),
            http_client=httpx.Client(proxies=os.getenv("OPENAI_PROXY", ""))
            )  # type: ignore
        return cls(
            generator_llm=generator_llm,
            critic_llm=critic_llm,
            embeddings_model=embeddings_model,
            chat_qa=chat_qa,
            chunk_size=chunk_size,
            testset_distribution=testset_distribution,
        )

    def _get_evolve_type(self) -> str:
        """
        Decides question evolution type based on probability
        """
        prob = self.rng.uniform(0, 1)
        return next(
            (
                key
                for key in self.testset_distribution.keys()
                if prob <= self.testset_distribution[key]
            ),
            "simple",
        )

    def _filter_context(self, context: str) -> bool:
        """
        context: str
            The input context

        Checks if the context is has information worthy of framing a question
        """
        human_prompt = SCORE_CONTEXT.format(context=context)
        prompt = ChatPromptTemplate.from_messages([human_prompt])
        results = self.critic_llm.generate(prompts=[prompt])
        output = results.generations[0][0].text.strip()
        score = load_as_score(output)
        return score >= self.threshold

    def _seed_question(self, context: str) -> str:
        human_prompt = SEED_QUESTION.format(context=context)
        prompt = ChatPromptTemplate.from_messages([human_prompt])
        results = self.generator_llm.generate(prompts=[prompt])
        return results.generations[0][0].text.strip()

    def _filter_question(self, question: str) -> bool:
        human_prompt = FILTER_QUESTION.format(question=question)
        prompt = ChatPromptTemplate.from_messages([human_prompt])

        results = self.critic_llm.generate(prompts=[prompt])
        results = results.generations[0][0].text.strip()
        json_results = load_as_json(results)
        return json_results.get("verdict") != "No"

    def _reasoning_question(self, question: str, context: str) -> str:
        return self._qc_template(REASONING_QUESTION, question, context)

    def _condition_question(self, question: str, context: str) -> str:
        return self._qc_template(CONDITIONAL_QUESTION, question, context)

    def _multicontext_question(
        self, question: str, context1: str, context2: str
    ) -> str:
        human_prompt = MULTICONTEXT_QUESTION.format(
            question=question, context1=context1, context2=context2
        )
        prompt = ChatPromptTemplate.from_messages([human_prompt])
        results = self.generator_llm.generate(prompts=[prompt])
        return results.generations[0][0].text.strip()

    def _compress_question(self, question: str) -> str:
        return self._question_transformation(COMPRESS_QUESTION, question=question)

    def _conversational_question(self, question: str) -> str:
        return self._question_transformation(CONVERSATION_QUESTION, question=question)

    def _question_transformation(self, prompt, question: str) -> str:
        human_prompt = prompt.format(question=question)
        prompt = ChatPromptTemplate.from_messages([human_prompt])
        results = self.generator_llm.generate(prompts=[prompt])
        return results.generations[0][0].text.strip()

    def _qc_template(self, prompt, question, context) -> str:
        human_prompt = prompt.format(question=question, context=context)
        prompt = ChatPromptTemplate.from_messages([human_prompt])
        results = self.generator_llm.generate(prompts=[prompt])
        return results.generations[0][0].text.strip()

    def _generate_answer(self, question: str, context: t.List[str]) -> t.List[str]:
        return [
            self._qc_template(ANSWER_FORMULATE, qstn, context[i])
            for i, qstn in enumerate(question.split("\n"))
        ]

    def _generate_context(self, question: str, text_chunk: str) -> t.List[str]:
        return [
            self._qc_template(CONTEXT_FORMULATE, qstn, text_chunk)
            for qstn in question.split("\n")
        ]

    def _remove_nodes(
        self, available_indices: t.List[BaseNode], node_idx: t.List
    ) -> t.List[BaseNode]:
        for idx in node_idx:
            available_indices.remove(idx)
        return available_indices

    def _generate_doc_nodes_map(
        self, document_nodes: t.List[BaseNode]
    ) -> t.Dict[str, t.List[BaseNode]]:
        doc_nodes_map: t.Dict[str, t.List[BaseNode]] = defaultdict(list)
        for node in document_nodes:
            if node.ref_doc_id:
                doc_nodes_map[node.ref_doc_id].append(node)

        return doc_nodes_map  # type: ignore

    def _get_neighbour_node(
        self, node: BaseNode, related_nodes: t.List[BaseNode]
    ) -> t.List[BaseNode]:
        if len(related_nodes) < 2:
            warnings.warn("No neighbors exists")
            return [node]
        idx = related_nodes.index(node)
        ids = [idx - 1, idx] if idx == (len(related_nodes) - 1) else [idx, idx + 1]
        return [related_nodes[idx] for idx in ids]

    def _embed_nodes(self, nodes: t.List[BaseNode]) -> t.Dict[str, t.List[float]]:
        embeddings = {}
        for node in nodes:
            embeddings[node.id_] = list(
                self.embedding_model.embed_query(node.get_content())
            )

        return embeddings

    def generate(
        self,
        documents: t.List[LlamaindexDocument] | t.List[LangchainDocument],
        test_size: int,
    ) -> TestDataset:
        if not isinstance(documents[0], (LlamaindexDocument, LangchainDocument)):
            raise ValueError(
                "Testset Generatation only supports LlamaindexDocuments or LangchainDocuments"  # noqa
            )

        if isinstance(documents[0], LangchainDocument):
            # cast to LangchainDocument since its the only case here
            documents = t.cast(t.List[LangchainDocument], documents)
            documents = [
                LlamaindexDocument.from_langchain_format(doc) for doc in documents
            ]
        # Convert documents into nodes
        node_parser = SimpleNodeParser.from_defaults(
            chunk_size=self.chunk_size, chunk_overlap=0, include_metadata=True
        )
        documents = t.cast(t.List[LlamaindexDocument], documents)
        document_nodes: t.List[BaseNode] = node_parser.get_nodes_from_documents(
            documents=documents
        )
        # maximum 1 seed question per node
        if test_size > len(document_nodes):
            raise ValueError(
                """Maximum possible number of samples exceeded, 
                             reduce test_size or add more documents"""
            )

        available_nodes = document_nodes
        doc_nodes_map = self._generate_doc_nodes_map(document_nodes)
        count_neighbours = sum(len(val) > 1 for _, val in doc_nodes_map.items())
        if count_neighbours < len(documents) // 2:
            warnings.warn("Most documents are too short")

        count = 0
        samples = []

        pbar = tqdm(total=test_size)
        while count < test_size and available_nodes != []:
            evolve_type = self._get_evolve_type()
            curr_node = self.rng.choice(np.array(available_nodes), size=1)[0]
            available_nodes = self._remove_nodes(available_nodes, [curr_node])

            neighbor_nodes = doc_nodes_map[curr_node.source_node.node_id]

            # Append multiple nodes randomly to remove chunking bias
            size = self.rng.integers(1, 3)
            nodes = (
                self._get_neighbour_node(curr_node, neighbor_nodes)
                if size > 1 and evolve_type != "multi_context"
                else [curr_node]
            )

            text_chunk = " ".join([node.get_content() for node in nodes])
            score = self._filter_context(text_chunk)
            if not score:
                continue
            seed_question = self._seed_question(text_chunk)
            is_valid_question = self._filter_question(seed_question)
            if not is_valid_question:
                continue

            if evolve_type == "multi_context":
                # Find most similar chunk in same document
                node_embedding = self._embed_nodes([nodes[-1]])
                neighbor_nodes = self._remove_nodes(neighbor_nodes, nodes)
                neighbor_emb = self._embed_nodes(neighbor_nodes)

                _, indices = get_top_k_embeddings(
                    list(node_embedding.values())[0],
                    list(neighbor_emb.values()),
                    similarity_cutoff=self.threshold / 10,
                )
                if indices:
                    # type cast indices from list[Any] to list[int]
                    indices = t.cast(t.List[int], indices)
                    best_neighbor = neighbor_nodes[indices[0]]
                    question = self._multicontext_question(
                        question=seed_question,
                        context1=text_chunk,
                        context2=best_neighbor.get_content(),
                    )
                    text_chunk = "\n".join([text_chunk, best_neighbor.get_content()])
                else:
                    continue

            # for reasoning and conditional modes, evolve question with the
            # functions from question_deep_map
            else:
                evolve_fun = question_deep_map.get(evolve_type)
                question = (
                    getattr(self, evolve_fun)(seed_question, text_chunk)
                    if evolve_fun
                    else seed_question
                )

            # compress question or convert into conversational questions
            if evolve_type != "simple":
                prob = self.rng.uniform(0, 1)
                if self.chat_qa and prob <= self.chat_qa:
                    question = self._conversational_question(question=question)
                else:
                    question = self._compress_question(question=question)

            is_valid_question = self._filter_question(question)
            if is_valid_question:
                context = self._generate_context(question, text_chunk)
                is_conv = len(context) > 1
                answer = self._generate_answer(question, context)
                for i, (qstn, ctx, ans) in enumerate(
                    zip(question.split("\n"), context, answer)
                ):
                    episode_done = False if is_conv and i == 0 else True
                    samples.append(
                        DataRow(qstn, [ctx], [ans], evolve_type, episode_done)
                    )
                count += 1
                pbar.update(count)

        return TestDataset(test_data=samples)



from centricube_ragas.metrics._answer_correctness import AnswerCorrectness, answer_correctness
from centricube_ragas.metrics._answer_relevance import AnswerRelevancy, answer_relevancy
from centricube_ragas.metrics._answer_similarity import AnswerSimilarity, answer_similarity
from centricube_ragas.metrics._context_precision import ContextPrecision, context_precision
from centricube_ragas.metrics._context_recall import ContextRecall, context_recall
from centricube_ragas.metrics._context_relevancy import ContextRelevancy, context_relevancy
from centricube_ragas.metrics._faithfulness import Faithfulness, faithfulness
from centricube_ragas.metrics.critique import AspectCritique
from centricube_ragas.metrics._answer_correctness_centricube import AnswerCorrectnessCentricube, answer_correctness_centricube
from centricube_ragas.metrics._answer_recall_centricube import AnswerRecallCentricube, answer_recall_centricube


DEFAULT_METRICS = [
    answer_relevancy,
    context_precision,
    faithfulness,
    context_recall,
    context_relevancy,
]

__all__ = [
    "Faithfulness",
    "faithfulness",
    "AnswerRelevancy",
    "answer_relevancy",
    "AnswerSimilarity",
    "answer_similarity",
    "AnswerCorrectness",
    "answer_correctness",
    "ContextRelevancy",
    "context_relevancy",
    "ContextPrecision",
    "context_precision",
    "AspectCritique",
    "ContextRecall",
    "context_recall",
    "AnswerCorrectnessCentricube",
    "answer_correctness_centricube",
    "AnswerRecallCentricube",
    "answer_recall_centricube"
]



from langchain.prompts import HumanMessagePromptTemplate

SEED_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Your task is to formulate a question from given context satisfying the rules given below:
    1.The question should make sense to humans even when read without the given context.
    2.The question should be fully answered from the given context.
    3.The question should be framed from a part of context that contains important information. It can also be from tables,code,etc.
    4.The answer to the question should not contain any links.
    5.The question should be of moderate difficulty.
    6.The question must be reasonable and must be understood and responded by humans.
    7.Do no use phrases like 'provided context',etc in the question
    8.Avoid framing question using word "and" that can be decomposed into more than one question.
    9.The question should not contain more than 10 words, make of use of abbreviation wherever possible.
    
context:{context}
"""  # noqa: E501
)


REASONING_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
You are a prompt rewriter. You will be provided with a question and a long context.Your task to is to complicate the given question to improve the difficulty of answering. 
You should do complicate the question by rewriting question into a multi-hop reasoning question based on the provided context. The question should require the reader to make multiple logical connections or inferences using the information available in given context. 
Here are some strategies to create multi-hop questions:

   - Bridge related entities: Identify information that relates specific entities and frame question that can be answered only by analysing information of both entities.
   
   - Use Pronouns: identify (he, she, it, they) that refer to same entity or concepts in the context, and ask questions that would require the reader to figure out what pronouns refer to.

   - Refer to Specific Details: Mention specific details or facts from different parts of the context including tables, code, etc and ask how they are related.

   - Pose Hypothetical Scenarios: Present a hypothetical situation or scenario that requires combining different elements from the context to arrive at an answer.

Rules to follow when rewriting question:
1. Ensure that the rewritten question can be answered entirely from the information present in the contexts.
2. Do not frame questions that contains more than 15 words. Use abbreviation wherever possible.
3. Make sure the question is clear and unambiguous. 
4. phrases like 'based on the provided context','according to the context',etc are not allowed to appear in the question.

question: {question}
CONTEXTS:
{context}

Multi-hop Reasoning Question:
"""  # noqa: E501
)

MULTICONTEXT_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
You are a prompt rewriter. You will be provided with a question and two set of contexts namely context1 and context2. 
Your task is to complicate the given question in a way that answering it requires information derived from both context1 and context2. 
Follow the rules given below while rewriting the question.
    1. The rewritten question should not be very long. Use abbreviation wherever possible.
    2. The rewritten question must be reasonable and must be understood and responded by humans.
    3. The rewritten question must be fully answerable from information present in context1 and context2. 
    4. Read and understand both contexts and rewrite the question so that answering requires insight from both context1 and context2.
    5. phrases like 'based on the provided context','according to the context?',etc are not allowed to appear in the question.

question:\n{question}
context1:\n{context1}
context2:\n{context2}
"""  # noqa: E501
)


CONDITIONAL_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Rewrite the provided question to increase its complexity by introducing a conditional element.
The goal is to make the question more intricate by incorporating a scenario or condition that affects the context of the question.
Follow the rules given below while rewriting the question.
    1. The rewritten question should not be longer than 25 words. Use abbreviation wherever possible.
    2. The rewritten question must be reasonable and must be understood and responded by humans.
    3. The rewritten question must be fully answerable from information present context.
    4. phrases like 'provided context','according to the context?',etc are not allowed to appear in the question.
for example,
question: What are the general principles for designing prompts in LLMs?
Rewritten Question:how to apply prompt designing principles to improve LLMs performance in reasoning tasks

question:{question}
context:\n{context}
Rewritten Question
"""  # noqa: E501
)


COMPRESS_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Rewrite the following question to make it more indirect and shorter while retaining the essence of the original question. The goal is to create a question that conveys the same meaning but in a less direct manner.
The rewritten question should shorter so use abbreviation wherever possible.
Original Question:
{question}

Indirectly Rewritten Question:
"""  # noqa: E501
)


CONVERSATION_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Reformat the provided question into two separate questions as if it were to be part of a conversation. Each question should focus on a specific aspect or subtopic related to the original question.
question: What are the advantages and disadvantages of remote work?
Reformatted Questions for Conversation: What are the benefits of remote work?\nOn the flip side, what challenges are encountered when working remotely?
question:{question}

Reformatted Questions for Conversation:
"""  # noqa: E501
)

SCORE_CONTEXT = HumanMessagePromptTemplate.from_template(
    """Evaluate the provided context and assign a numerical score between 0 and 10 based on the following criteria:
1. Award a high score to context that thoroughly delves into and explains concepts.
2. Assign a lower score to context that contains excessive references, acknowledgments, external links, personal information, or other non-essential elements.
Output the score only.
Context:
{context}
Score:
"""  # noqa: E501
)

FILTER_QUESTION = HumanMessagePromptTemplate.from_template(
    """\
Determine if the given question can be clearly understood even when presented without any additional context. Specify reason and verdict is a valid json format.
question: What is the keyword that best describes the paper's focus in natural language understanding tasks?
{{"reason":"The specific paper being referred to is not mentioned in the question.", "verdict": "No"}}
question:{question}
"""  # noqa: E501
)


ANSWER_FORMULATE = HumanMessagePromptTemplate.from_template(
    """\
Answer the question using the information from the given context. 
question:{question}
context:{context}
answer:
"""  # noqa: E501
)

CONTEXT_FORMULATE = HumanMessagePromptTemplate.from_template(
    """Please extract relevant sentences from the provided context that can potentially help answer the following question. While extracting candidate sentences you're not allowed to make any changes to sentences from given context.

question:{question}
context:\n{context}
candidate sentences:\n
"""  # noqa: E501
)



from __future__ import annotations

import typing as t
from dataclasses import dataclass

import numpy as np
from datasets import Dataset
from langchain.callbacks.manager import CallbackManager, trace_as_chain_group
from langchain.prompts import ChatPromptTemplate, HumanMessagePromptTemplate

from centricube_ragas.metrics.base import EvaluationMode, MetricWithLLM
from centricube_ragas.utils import json_loader

if t.TYPE_CHECKING:
    from langchain.callbacks.base import Callbacks

CONTEXT_PRECISION = HumanMessagePromptTemplate.from_template(
    """\
Verify if the information in the given context is useful in answering the question.

question: What are the health benefits of green tea?
context: 
This article explores the rich history of tea cultivation in China, tracing its roots back to the ancient dynasties. It discusses how different regions have developed their unique tea varieties and brewing techniques. The article also delves into the cultural significance of tea in Chinese society and how it has become a symbol of hospitality and relaxation.
verification:
{{"reason":"The context, while informative about the history and cultural significance of tea in China, does not provide specific information about the health benefits of green tea. Thus, it is not useful for answering the question about health benefits.", "verdict":"No"}}

question: How does photosynthesis work in plants?
context:
Photosynthesis in plants is a complex process involving multiple steps. This paper details how chlorophyll within the chloroplasts absorbs sunlight, which then drives the chemical reaction converting carbon dioxide and water into glucose and oxygen. It explains the role of light and dark reactions and how ATP and NADPH are produced during these processes.
verification:
{{"reason":"This context is extremely relevant and useful for answering the question. It directly addresses the mechanisms of photosynthesis, explaining the key components and processes involved.", "verdict":"Yes"}}

question:{question}
context:
{context}
verification:"""  # noqa: E501
)


@dataclass
class ContextPrecision(MetricWithLLM):
    """
    Average Precision is a metric that evaluates whether all of the
    relevant items selected by the model are ranked higher or not.

    Attributes
    ----------
    name : str
    batch_size : int
        Batch size for openai completion.
    """

    name: str = "context_precision"  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.qc  # type: ignore
    batch_size: int = 15

    def _score_batch(
        self: t.Self,
        dataset: Dataset,
        callbacks: t.Optional[Callbacks] = None,
        callback_group_name: str = "batch",
    ) -> list:
        prompts = []
        questions, contexts = dataset["question"], dataset["contexts"]

        cb = CallbackManager.configure(inheritable_callbacks=callbacks)
        with trace_as_chain_group(
            callback_group_name, callback_manager=cb
        ) as batch_group:
            for qstn, ctx in zip(questions, contexts):
                human_prompts = [
                    ChatPromptTemplate.from_messages(
                        [CONTEXT_PRECISION.format(question=qstn, context=c)]
                    )
                    for c in ctx
                ]

                prompts.extend(human_prompts)

            responses: list[list[str]] = []
            results = self.llm.generate(
                prompts,
                n=1,
                callbacks=batch_group,
            )
            responses = [[i.text for i in r] for r in results.generations]
            context_lens = [len(ctx) for ctx in contexts]
            context_lens.insert(0, 0)
            context_lens = np.cumsum(context_lens)
            grouped_responses = [
                responses[start:end]
                for start, end in zip(context_lens[:-1], context_lens[1:])
            ]
            scores = []

            for response in grouped_responses:
                response = [
                    json_loader.safe_load(item, self.llm) for item in sum(response, [])
                ]
                response = [
                    int("yes" in resp.get("verdict", " ").lower())
                    if resp.get("verdict")
                    else np.nan
                    for resp in response
                ]
                denominator = sum(response) + 1e-10
                numerator = sum(
                    [
                        (sum(response[: i + 1]) / (i + 1)) * response[i]
                        for i in range(len(response))
                    ]
                )
                scores.append(numerator / denominator)

        return scores


context_precision = ContextPrecision()



from centricube_ragas.evaluation import evaluate

try:
    from ._version import version as __version__
    from ._version import version_tuple
except ImportError:
    __version__ = "unknown version"
    version_tuple = (0, 0, "unknown version")


__all__ = ["evaluate", "__version__", "version_tuple"]



from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod

from langchain.schema import LLMResult

if t.TYPE_CHECKING:
    from langchain.callbacks.base import Callbacks
    from langchain.prompts import ChatPromptTemplate


class RagasLLM(ABC):
    """
    BaseLLM is the base class for all LLMs. It provides a consistent interface for other
    classes that interact with LLMs like Langchains, LlamaIndex, LiteLLM etc. Handles
    multiple_completions even if not supported by the LLM.

    It currently takes in ChatPromptTemplates and returns LLMResults which are Langchain
    primitives.
    """

    # supports multiple compeletions for the given prompt
    n_completions_supported: bool = False

    @property
    @abstractmethod
    def llm(self) -> t.Any:
        ...

    def validate_api_key(self):
        """
        Validates that the api key is set for the LLM
        """
        pass

    @abstractmethod
    def generate(
        self,
        prompts: list[ChatPromptTemplate],
        n: int = 1,
        temperature: float = 1e-8,
        callbacks: t.Optional[Callbacks] = None,
    ) -> LLMResult:
        ...

    @abstractmethod
    async def agenerate(
        self,
        prompt: ChatPromptTemplate,
        n: int = 1,
        temperature: float = 1e-8,
        callbacks: t.Optional[Callbacks] = None,
    ) -> LLMResult:
        ...



from __future__ import annotations

import typing as t
from dataclasses import dataclass, field

import numpy as np
from datasets import Dataset
from langchain.callbacks.manager import CallbackManager, trace_as_chain_group
from langchain_community.embeddings import OpenAIEmbeddings
from langchain.prompts import ChatPromptTemplate, HumanMessagePromptTemplate

from centricube_ragas.embeddings.base import embedding_factory
from centricube_ragas.exceptions import OpenAIKeyNotFound
from centricube_ragas.metrics.base import EvaluationMode, MetricWithLLM
from centricube_ragas.utils import json_loader

if t.TYPE_CHECKING:
    from langchain.callbacks.base import Callbacks

    from centricube_ragas.embeddings.base import RagasEmbeddings


QUESTION_GEN = HumanMessagePromptTemplate.from_template(
    """
Generate a question for the given answer and Identify if answer is noncommittal

Answer:
Albert Einstein was born in Germany.
Context:
Albert Einstein was a German-born theoretical physicist who is widely held to be one of the greatest and most influential scientists of all time
Output:
{{"question":"Where was Albert Einstein born?","noncommittal":false}}


Answer:
It can change its skin color based on the temperature of its environment.
Context:
A recent scientific study has discovered a new species of frog in the Amazon rainforest that has the unique ability to change its skin color based on the temperature of its environment.
Output:
{{"question":"What unique ability does the newly discovered species of frog have?","noncommittal":false}}


Answer:
Everest
Context:
The tallest mountain on Earth, measured from sea level, is a renowned peak located in the Himalayas.
Output:
{{"question":"What is the tallest mountain on Earth?","noncommittal":false}}


Answer:
I don't know about the  groundbreaking feature of the smartphone invented in 2023 as am unware of information beyong 2022. 
Context:
In 2023, a groundbreaking invention was announced: a smartphone with a battery life of one month, revolutionizing the way people use mobile technology.
Output:
{{"question":"What was the groundbreaking feature of the smartphone invented in 2023?", "noncommittal":true}}



Answer:
{answer}
Context:
{context}
Output:"""  # noqa: E501
)


@dataclass
class AnswerRelevancy(MetricWithLLM):
    """
    Scores the relevancy of the answer according to the given question.
    Answers with incomplete, redundant or unnecessary information is penalized.
    Score can range from 0 to 1 with 1 being the best.

    Attributes
    ----------
    name: string
        The name of the metrics
    batch_size: int
        batch size for evaluation
    strictness: int
        Here indicates the number questions generated per answer.
        Ideal range between 3 to 5.
    embeddings: Embedding
        The langchain wrapper of Embedding object.
        E.g. HuggingFaceEmbeddings('BAAI/bge-base-en')
    """

    name: str = "answer_relevancy"  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.qac  # type: ignore
    batch_size: int = 15
    strictness: int = 3
    embeddings: RagasEmbeddings = field(default_factory=embedding_factory)

    def init_model(self):
        super().init_model()

        if isinstance(self.embeddings, OpenAIEmbeddings):
            if self.embeddings.openai_api_key == "no-key":
                raise OpenAIKeyNotFound

    def _score_batch(
        self: t.Self,
        dataset: Dataset,
        callbacks: t.Optional[Callbacks] = None,
        callback_group_name: str = "batch",
    ) -> list[float]:
        questions, answers, contexts = (
            dataset["question"],
            dataset["answer"],
            dataset["contexts"],
        )

        cb = CallbackManager.configure(inheritable_callbacks=callbacks)
        with trace_as_chain_group(
            callback_group_name, callback_manager=cb
        ) as batch_group:
            prompts = []
            for ans, ctx in zip(answers, contexts):
                human_prompt = QUESTION_GEN.format(answer=ans, context="\n".join(ctx))
                prompts.append(ChatPromptTemplate.from_messages([human_prompt]))

            results = self.llm.generate(
                prompts,
                n=self.strictness,
                callbacks=batch_group,
            )
            results = [
                [json_loader.safe_load(i.text, self.llm) for i in r]
                for r in results.generations
            ]
            scores = []
            for question, result in zip(questions, results):
                gen_questions = [item.get("question", "") for item in result]
                committal = np.any([item.get("noncommittal", False) for item in result])
                cosine_sim = self.calculate_similarity(question, gen_questions)
                scores.append(cosine_sim.mean() * int(not committal))

        return scores

    def calculate_similarity(
        self: t.Self, question: str, generated_questions: list[str]
    ):
        assert self.embeddings is not None
        question_vec = np.asarray(self.embeddings.embed_query(question)).reshape(1, -1)
        gen_question_vec = np.asarray(
            self.embeddings.embed_documents(generated_questions)
        )
        norm = np.linalg.norm(gen_question_vec, axis=1) * np.linalg.norm(
            question_vec, axis=1
        )
        return (
            np.dot(gen_question_vec, question_vec.T).reshape(
                -1,
            )
            / norm
        )


answer_relevancy = AnswerRelevancy()



from __future__ import annotations

import typing as t
from dataclasses import dataclass, field

import numpy as np
from datasets import Dataset

from centricube_ragas.embeddings.base import (
    HuggingfaceEmbeddings,
    OpenAIEmbeddings,
    embedding_factory,
)
from centricube_ragas.exceptions import OpenAIKeyNotFound
from centricube_ragas.metrics.base import EvaluationMode, MetricWithLLM

if t.TYPE_CHECKING:
    from langchain.callbacks.base import Callbacks

    from centricube_ragas.embeddings.base import RagasEmbeddings


@dataclass
class AnswerSimilarity(MetricWithLLM):
    """
    Scores the semantic similarity of ground truth with generated answer.
    cross encoder score is used to quantify semantic similarity.
    SAS paper: https://arxiv.org/pdf/2108.06130.pdf

    Attributes
    ----------
    name : str
    batch_size : int
        Batch size.
    model_name:
        The model to be used for calculating semantic similarity
        Defaults open-ai-embeddings
        select cross-encoder model for best results
        https://huggingface.co/spaces/mteb/leaderboard
    threshold:
        The threshold if given used to map output to binary
        Default 0.5
    """

    name: str = "answer_similarity"  # type: ignore
    evaluation_mode: EvaluationMode = EvaluationMode.ga  # type: ignore
    batch_size: int = 15
    embeddings: RagasEmbeddings = field(default_factory=embedding_factory)
    is_cross_encoder: bool = False
    threshold: t.Optional[float] = None

    def __post_init__(self: t.Self):
        # only for cross encoder
        if isinstance(self.embeddings, HuggingfaceEmbeddings):
            self.is_cross_encoder = True if self.embeddings.is_cross_encoder else False
            self.embeddings.encode_kwargs = {
                "batch_size": self.batch_size,
            }

    def init_model(self):
        super().init_model()

        if isinstance(self.embeddings, OpenAIEmbeddings):
            if self.embeddings.openai_api_key == "no-key":
                raise OpenAIKeyNotFound

    def _score_batch(
        self: t.Self,
        dataset: Dataset,
        callbacks: t.Optional[Callbacks] = None,
        callback_group_name: str = "batch",
    ) -> list[float]:
        ground_truths, answers = dataset["ground_truths"], dataset["answer"]
        ground_truths = [item[0] for item in ground_truths]

        if self.is_cross_encoder:
            assert isinstance(self.embeddings, HuggingfaceEmbeddings)
            inputs = [list(item) for item in list(zip(ground_truths, answers))]
            scores = np.array(self.embeddings.predict(inputs))
        else:
            embeddings_1 = np.array(self.embeddings.embed_documents(ground_truths))
            embeddings_2 = np.array(self.embeddings.embed_documents(answers))
            similarity = embeddings_1 @ embeddings_2.T
            scores = np.diagonal(similarity)

        assert isinstance(scores, np.ndarray), "Expects ndarray"
        if self.threshold:
            scores = scores >= self.threshold  # type: ignore

        return scores.tolist()


answer_similarity = AnswerSimilarity()



import io
import os

import setuptools


def get_version():
    if os.getenv('RELEASE_VERSION'):
        version = os.environ['RELEASE_VERSION']
    else:
        version_file = os.path.join(os.path.dirname(__file__), 'version.txt')
        with open(version_file, 'r') as f:
            version = f.read().strip()
    return version.lstrip('v')


def read_requirements_file(filepath):
    with open(filepath) as fin:
        requirements = fin.read()
    return requirements


extras = {}
REQUIRED_PACKAGES = read_requirements_file('requirements.txt')


def read(*names, **kwargs):
    with io.open(os.path.join(os.path.dirname(__file__), *names), encoding=kwargs.get('encoding', 'utf8')) as fp:
        return fp.read()


def get_package_data_files(package, data, package_dir=None):
    """
    Helps to list all specified files in package including files in directories
    since `package_data` ignores directories.
    """
    if package_dir is None:
        package_dir = os.path.join(*package.split('.'))
    all_files = []
    for f in data:
        path = os.path.join(package_dir, f)
        if os.path.isfile(path):
            all_files.append(f)
            continue
        for root, _dirs, files in os.walk(path, followlinks=True):
            root = os.path.relpath(root, package_dir)
            for file in files:
                file = os.path.join(root, file)
                if file not in all_files:
                    all_files.append(file)
    return all_files


setuptools.setup(
    name='centricube_ragas',
    version=get_version(),
    author='CICIT',
    author_email='jweelee@qq.com',
    description='centricube ragas modules',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/cicit/centricube',
    packages=setuptools.find_packages(exclude=('examples*', 'tests*', 'applications*', 'model_zoo*'),),
    package_data={},
    setup_requires=[],
    install_requires=REQUIRED_PACKAGES,
    entry_points={},
    extras_require=extras,
    python_requires='>=3.6',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    license='Apache 2.0',
)



"""
Q - question
A - answer: generated_text from RAG pipeline
C - contexts: context used for generation
G - ground_truths: ground truth answer
"""
from __future__ import annotations

import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from math import floor

from datasets import Dataset
from centricube_ragas.embeddings.base import RagasEmbeddings
from centricube_ragas.llms import llm_factory
from tqdm import tqdm

from langchain.callbacks.manager import CallbackManager, trace_as_chain_group

if t.TYPE_CHECKING:
    from centricube_ragas.llms import RagasLLM

    from langchain.callbacks.base import Callbacks


def make_batches(total_size: int, batch_size: int) -> list[range]:
    """
    Take a total size and batch size and return a list of ranges for the batches
    """
    tail = total_size % batch_size
    num_batches = floor(total_size / batch_size)
    batches = [
        range(i, i + batch_size) for i in range(0, batch_size * num_batches, batch_size)
    ]
    if tail != 0:
        batches.append(range(batch_size * num_batches, batch_size * num_batches + tail))

    return batches


EvaluationMode = Enum("EvaluationMode", "qac qa qc gc ga qga qcg")


@dataclass
class Metric(ABC):
    batch_size: int

    @property
    @abstractmethod
    def name(self) -> t.Union[str, tuple]:
        ...

    @property
    @abstractmethod
    def evaluation_mode(self) -> EvaluationMode:
        ...

    @abstractmethod
    def init_model(self):
        """
        This method will lazy initialize the model.
        """
        ...

    def score(
        self: t.Self,
        dataset: Dataset,
        callbacks: t.Optional[Callbacks] = None,
    ) -> Dataset:
        scores = []
        cm = CallbackManager.configure(inheritable_callbacks=callbacks)
        with trace_as_chain_group(f"ragas_{self.name}", callback_manager=cm) as group:
            for batch in tqdm(self.get_batches(len(dataset))):
                score = self._score_batch(dataset.select(batch), callbacks=group)
                scores.extend(score)
        
        if isinstance(self.name, tuple):
            for i, name in enumerate(self.name):
                dataset = dataset.add_column(f"{name}", [s[i] for s in scores])
        else:
            dataset = dataset.add_column(f"{self.name}", scores)

        return dataset

    @abstractmethod
    def _score_batch(
        selfself: t.Self,
        dataset: Dataset,
        callbacks: t.Optional[Callbacks] = None,
        callback_group_name: str = "batch",
    ) -> list:
        ...

    def score_single(
        self: t.Self,
        ds_row: dict,
        callbacks: t.Optional[Callbacks] = None,
    ) -> float:
        """
        Score for a single row of dataset
        """
        # TODO: validation check if they are string

        ds = Dataset.from_dict({k: [v] for k, v in ds_row.items()})
        score = self._score_batch(
            ds, callback_group_name=self.name, callbacks=callbacks
        )

        return score[0]

    def get_batches(self, dataset_size: int) -> list[range]:
        return make_batches(dataset_size, self.batch_size)


@dataclass
class MetricWithLLM(Metric):
    llm: RagasLLM = field(default_factory=llm_factory)

    def init_model(self):
        """
        Init any models in the metric, this is invoked before evaluate()
        to load all the models
        Also check if the api key is valid for OpenAI and AzureOpenAI
        """
        if hasattr(self.llm, "validate_api_key"):
            self.llm.validate_api_key()
        if hasattr(self, "embeddings"):
            # since we are using Langchain Embeddings directly, we need to check this
            if hasattr(self.embeddings, "validate_api_key"):
                self.embeddings = t.cast(RagasEmbeddings, self.embeddings)
                self.embeddings.validate_api_key()
